"""Provide the current Flet version."""

import os
import subprocess as sp
from pathlib import Path

import flet
from flet.utils import is_mobile, is_windows, which

# will be replaced by CI
version = "0.28.3"


def update_version():
    """Return the current version or default."""
    working = Path().absolute()
    os.chdir(Path(flet.__file__).absolute().parent)
    in_repo = (
        which("git.exe" if is_windows() else "git")
        and sp.run(
            ["git", "status"],
            capture_output=True,
            text=True,
        ).returncode
        == 0
    )

    if in_repo:
        # NOTE: this may break if there is a tag name starting with
        #         "v" that isn't a version number
        class RepositoryError(OSError):
            pass

        git_p = sp.run(
            ["git", "describe", "--abbrev=0"],
            capture_output=True,
            text=True,
        )
        err = git_p.stderr.strip()

        if "cannot describe anything" in err:
            msg = "You may be using a repo cloned from a fork. "
            msg += "If so please clone the original Flet repo"
            raise RepositoryError(msg)

        if err:
            msg = "Unknown error while fetching the version: {err}"
            raise RepositoryError(msg)
        version = git_p.stdout.strip()[1:]

    else:
        version = "0.2.0"
    os.chdir(working)
    return version


if not version and not is_mobile():
    version = update_version()



from flet.auth.authorization import Authorization
from flet.auth.group import Group
from flet.auth.oauth_provider import OAuthProvider
from flet.auth.oauth_token import OAuthToken
from flet.auth.user import User



from typing import Any, Optional

from flet.core.control import Control
from flet.core.ref import Ref
from flet.utils import deprecated


@deprecated(
    reason="Flashlight control has been moved to a separate Python package: https://pypi.org/project/flet-flashlight. "
    + "Read more about this change in Flet blog: https://flet.dev/blog/flet-v-0-26-release-announcement",
    version="0.26.0",
    delete_version="0.29.0",
)
class Flashlight(Control):
    """
    A control to use FlashLight. Works on iOS and Android. Based on torch_light Flutter widget (https://pub.dev/packages/torch_light).

    Flashlight control is non-visual and should be added to `page.overlay` list.

    Example:
    ```
    import flet as ft

    def main(page: ft.Page):
        flashLight = ft.Flashlight()
        page.overlay.append(flashLight)
        page.add(
            ft.TextButton("toggle", on_click: lambda _: flashlight.toggle())
        )

    ft.app(target=main)
    ```

    """

    def __init__(
        self,
        ref: Optional[Ref] = None,
        data: Any = None,
    ):
        Control.__init__(
            self,
            ref=ref,
            data=data,
        )

        self.turned_on = False

    def _get_control_name(self):
        return "flashlight"

    def turn_on(self, wait_timeout: Optional[int] = 5) -> bool:
        sr = self.invoke_method("on", wait_for_result=True, wait_timeout=wait_timeout)

        if int(sr) == 1:
            self.turned_on = True
        return self.turned_on

    async def turn_on_async(self, wait_timeout: Optional[int] = 5) -> bool:
        sr = await self.invoke_method_async(
            "on", wait_for_result=True, wait_timeout=wait_timeout
        )
        if int(sr) == 1:
            self.turned_on = True
        return self.turned_on

    def turn_off(self, wait_timeout: Optional[int] = 5) -> bool:
        sr = self.invoke_method("off", wait_for_result=True, wait_timeout=wait_timeout)

        if int(sr) == 1:
            self.turned_on = False
        return self.turned_on

    async def turn_off_async(self, wait_timeout: Optional[int] = 5) -> bool:
        sr = await self.invoke_method_async(
            "off", wait_for_result=True, wait_timeout=wait_timeout
        )
        if int(sr) == 1:
            self.turned_on = False
        return self.turned_on

    def toggle(self, wait_timeout: Optional[int] = 5) -> bool:
        if self.turned_on:
            return self.turn_off(wait_timeout)
        return self.turn_on(wait_timeout)

    async def toggle_async(self, wait_timeout: Optional[int] = 5) -> bool:
        if self.turned_on:
            return await self.turn_off_async(wait_timeout)
        return await self.turn_on_async(wait_timeout)



from flet.auth.oauth_provider import OAuthProvider


class GoogleOAuthProvider(OAuthProvider):
    def __init__(self, client_id: str, client_secret: str, redirect_url: str) -> None:
        super().__init__(
            client_id=client_id,
            client_secret=client_secret,
            authorization_endpoint="https://accounts.google.com/o/oauth2/auth",
            token_endpoint="https://oauth2.googleapis.com/token",
            redirect_url=redirect_url,
            user_scopes=[
                "https://www.googleapis.com/auth/userinfo.email",
                "https://www.googleapis.com/auth/userinfo.profile",
            ],
            user_endpoint="https://www.googleapis.com/oauth2/v3/userinfo",
            user_id_fn=lambda u: u["sub"],
            group_scopes=[],
        )



from dataclasses import dataclass
from typing import Optional

from flet.core.types import ColorValue, OptionalNumber


@dataclass
class ChartPointShape:
    pass


@dataclass
class ChartCirclePoint(ChartPointShape):
    color: Optional[ColorValue] = None
    radius: OptionalNumber = None
    stroke_color: Optional[ColorValue] = None
    stroke_width: OptionalNumber = None

    def __post_init__(self):
        self.type = "circle"


@dataclass
class ChartSquarePoint(ChartPointShape):
    color: Optional[ColorValue] = None
    size: OptionalNumber = None
    stroke_color: Optional[ColorValue] = None
    stroke_width: OptionalNumber = None

    def __post_init__(self):
        self.type = "square"


@dataclass
class ChartCrossPoint(ChartPointShape):
    color: Optional[ColorValue] = None
    size: OptionalNumber = None
    width: OptionalNumber = None

    def __post_init__(self):
        self.type = "cross"



from typing import Any, Optional, Union

from flet.core.animation import AnimationValue
from flet.core.badge import BadgeValue
from flet.core.constrained_control import ConstrainedControl
from flet.core.control import OptionalNumber
from flet.core.ref import Ref
from flet.core.tooltip import TooltipValue
from flet.core.types import (
    ColorEnums,
    ColorValue,
    LabelPosition,
    MouseCursor,
    OffsetValue,
    OptionalControlEventCallable,
    ResponsiveNumber,
    RotateValue,
    ScaleValue,
)

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class CupertinoRadio(ConstrainedControl):
    """
    Radio buttons let people select a single option from two or more choices.

    -----

    Online docs: https://flet.dev/docs/controls/cupertinoradio
    """

    def __init__(
        self,
        label: Optional[str] = None,
        value: Optional[str] = None,
        label_position: Optional[LabelPosition] = None,
        fill_color: Optional[ColorValue] = None,
        active_color: Optional[ColorValue] = None,
        inactive_color: Optional[ColorValue] = None,
        autofocus: Optional[bool] = None,
        use_checkmark_style: Optional[bool] = None,
        toggleable: Optional[bool] = None,
        focus_color: Optional[ColorValue] = None,
        mouse_cursor: Optional[MouseCursor] = None,
        on_focus: OptionalControlEventCallable = None,
        on_blur: OptionalControlEventCallable = None,
        #
        # ConstrainedControl
        #
        ref: Optional[Ref] = None,
        key: Optional[str] = None,
        width: OptionalNumber = None,
        height: OptionalNumber = None,
        left: OptionalNumber = None,
        top: OptionalNumber = None,
        right: OptionalNumber = None,
        bottom: OptionalNumber = None,
        expand: Union[None, bool, int] = None,
        expand_loose: Optional[bool] = None,
        col: Optional[ResponsiveNumber] = None,
        opacity: OptionalNumber = None,
        rotate: Optional[RotateValue] = None,
        scale: Optional[ScaleValue] = None,
        offset: Optional[OffsetValue] = None,
        aspect_ratio: OptionalNumber = None,
        animate_opacity: Optional[AnimationValue] = None,
        animate_size: Optional[AnimationValue] = None,
        animate_position: Optional[AnimationValue] = None,
        animate_rotation: Optional[AnimationValue] = None,
        animate_scale: Optional[AnimationValue] = None,
        animate_offset: Optional[AnimationValue] = None,
        on_animation_end: OptionalControlEventCallable = None,
        tooltip: Optional[TooltipValue] = None,
        badge: Optional[BadgeValue] = None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
    ):
        ConstrainedControl.__init__(
            self,
            ref=ref,
            key=key,
            width=width,
            height=height,
            left=left,
            top=top,
            right=right,
            bottom=bottom,
            expand=expand,
            expand_loose=expand_loose,
            col=col,
            opacity=opacity,
            rotate=rotate,
            scale=scale,
            offset=offset,
            aspect_ratio=aspect_ratio,
            animate_opacity=animate_opacity,
            animate_size=animate_size,
            animate_position=animate_position,
            animate_rotation=animate_rotation,
            animate_scale=animate_scale,
            animate_offset=animate_offset,
            on_animation_end=on_animation_end,
            tooltip=tooltip,
            badge=badge,
            visible=visible,
            disabled=disabled,
            data=data,
        )
        self.value = value
        self.label = label
        self.label_position = label_position
        self.autofocus = autofocus
        self.use_checkmark_style = use_checkmark_style
        self.fill_color = fill_color
        self.active_color = active_color
        self.inactive_color = inactive_color
        self.on_focus = on_focus
        self.on_blur = on_blur
        self.focus_color = focus_color
        self.toggleable = toggleable
        self.mouse_cursor = mouse_cursor

    def _get_control_name(self):
        return "cupertinoradio"

    # value
    @property
    def value(self) -> Optional[str]:
        return self._get_attr("value", def_value="")

    @value.setter
    def value(self, value: Optional[str]):
        self._set_attr("value", value)

    # label
    @property
    def label(self) -> Optional[str]:
        return self._get_attr("label")

    @label.setter
    def label(self, value: Optional[str]):
        self._set_attr("label", value)

    # label_position
    @property
    def label_position(self) -> Optional[LabelPosition]:
        return self.__label_position

    @label_position.setter
    def label_position(self, value: Optional[LabelPosition]):
        self.__label_position = value
        self._set_enum_attr("labelPosition", value, LabelPosition)

    # fill_color
    @property
    def fill_color(self) -> Optional[ColorValue]:
        return self.__fill_color

    @fill_color.setter
    def fill_color(self, value: Optional[ColorValue]):
        self.__fill_color = value
        self._set_enum_attr("fillColor", value, ColorEnums)

    # mouse_cursor
    @property
    def mouse_cursor(self) -> Optional[MouseCursor]:
        return self.__mouse_cursor

    @mouse_cursor.setter
    def mouse_cursor(self, value: Optional[MouseCursor]):
        self.__mouse_cursor = value
        self._set_enum_attr("mouseCursor", value, MouseCursor)

    # focus_color
    @property
    def focus_color(self) -> Optional[ColorValue]:
        return self.__focus_color

    @focus_color.setter
    def focus_color(self, value: Optional[ColorValue]):
        self.__focus_color = value
        self._set_enum_attr("focusColor", value, ColorEnums)

    # toggleable
    @property
    def toggleable(self) -> bool:
        return self._get_attr("toggleable", data_type="bool", def_value=False)

    @toggleable.setter
    def toggleable(self, value: Optional[bool]):
        self._set_attr("toggleable", value)

    # on_focus
    @property
    def on_focus(self) -> OptionalControlEventCallable:
        return self._get_event_handler("focus")

    @on_focus.setter
    def on_focus(self, handler: OptionalControlEventCallable):
        self._add_event_handler("focus", handler)

    # on_blur
    @property
    def on_blur(self) -> OptionalControlEventCallable:
        return self._get_event_handler("blur")

    @on_blur.setter
    def on_blur(self, handler: OptionalControlEventCallable):
        self._add_event_handler("blur", handler)

    # autofocus
    @property
    def autofocus(self) -> bool:
        return self._get_attr("autofocus", data_type="bool", def_value=False)

    @autofocus.setter
    def autofocus(self, value: Optional[bool]):
        self._set_attr("autofocus", value)

    # use_checkmark_style
    @property
    def use_checkmark_style(self) -> bool:
        return self._get_attr("useCheckmarkStyle", data_type="bool", def_value=False)

    @use_checkmark_style.setter
    def use_checkmark_style(self, value: Optional[bool]):
        self._set_attr("useCheckmarkStyle", value)

    # active_color
    @property
    def active_color(self) -> Optional[ColorValue]:
        return self.__active_color

    @active_color.setter
    def active_color(self, value: Optional[ColorValue]):
        self.__active_color = value
        self._set_enum_attr("activeColor", value, ColorEnums)

    # inactive_color
    @property
    def inactive_color(self) -> Optional[ColorValue]:
        return self.__inactive_color

    @inactive_color.setter
    def inactive_color(self, value: Optional[ColorValue]):
        self.__inactive_color = value
        self._set_enum_attr("inactiveColor", value, ColorEnums)



import time
import warnings
from typing import Any, List, Optional, Union

from flet.core.alignment import Alignment
from flet.core.animation import AnimationValue
from flet.core.badge import BadgeValue
from flet.core.box import BoxConstraints
from flet.core.control import Control, OptionalNumber
from flet.core.form_field_control import FormFieldControl, InputBorder
from flet.core.ref import Ref
from flet.core.text_style import TextStyle
from flet.core.tooltip import TooltipValue
from flet.core.types import (
    BorderRadiusValue,
    ColorEnums,
    ColorValue,
    DurationValue,
    IconEnums,
    IconValueOrControl,
    OffsetValue,
    OptionalControlEventCallable,
    PaddingValue,
    ResponsiveNumber,
    RotateValue,
    ScaleValue,
)


class Option(Control):
    def __init__(
        self,
        key: Optional[str] = None,
        text: Optional[str] = None,
        content: Optional[Control] = None,
        alignment: Optional[Alignment] = None,
        text_style: Optional[TextStyle] = None,
        on_click: OptionalControlEventCallable = None,
        #
        # Control
        #
        ref=None,
        disabled: Optional[bool] = None,
        visible: Optional[bool] = None,
        data: Any = None,
    ):
        Control.__init__(self, ref=ref, disabled=disabled, visible=visible, data=data)
        self.key = key
        self.text = text
        self.content = content
        self.on_click = on_click
        self.alignment = alignment
        self.text_style = text_style

    def _get_control_name(self):
        return "dropdownoption"

    def _get_children(self):
        children = []
        if self.__content is not None:
            self.__content._set_attr_internal("n", "content")
            children.append(self.__content)
        return children

    def before_update(self):
        super().before_update()
        assert (
            self.key is not None or self.text is not None
        ), "key or text must be specified"
        self._set_attr_json("alignment", self.__alignment)
        if isinstance(self.__text_style, TextStyle):
            self._set_attr_json("textStyle", self.__text_style)

    # key
    @property
    def key(self) -> Optional[str]:
        return self._get_attr("key")

    @key.setter
    def key(self, value: Optional[str]):
        self._set_attr("key", value)

    # text
    @property
    def text(self) -> Optional[str]:
        return self._get_attr("text")

    @text.setter
    def text(self, value: Optional[str]):
        self._set_attr("text", value)

    # content
    @property
    def content(self) -> Optional[Control]:
        return self.__content

    @content.setter
    def content(self, value: Optional[Control]):
        self.__content = value

    # alignment
    @property
    def alignment(self) -> Optional[Alignment]:
        return self.__alignment

    @alignment.setter
    def alignment(self, value: Optional[Alignment]):
        self.__alignment = value

    # text_style
    @property
    def text_style(self) -> Optional[TextStyle]:
        return self.__text_style

    @text_style.setter
    def text_style(self, value: Optional[TextStyle]):
        self.__text_style = value

    # on_click
    @property
    def on_click(self) -> OptionalControlEventCallable:
        return self._get_event_handler("click")

    @on_click.setter
    def on_click(self, handler: OptionalControlEventCallable):
        self._add_event_handler("click", handler)


class DropdownM2(FormFieldControl):
    """
    A dropdown lets the user select from a number of items. The dropdown shows the currently selected item as well as an arrow that opens a menu for selecting another item.

    Example:
    ```
    import flet as ft

    def main(page: ft.Page):
        def button_clicked(e):
            t.value = f"Dropdown value is:  {dd.value}"
            page.update()

        t = ft.Text()
        b = ft.ElevatedButton(text="Submit", on_click=button_clicked)
        dd = ft.DropdownM2(
            width=200,
            options=[
                ft.dropdown.Option("Red"),
                ft.dropdown.Option("Green"),
                ft.dropdown.Option("Blue"),
            ],
        )
        page.add(dd, b, t)

    ft.app(target=main)
    ```

    -----

    Online docs: https://flet.dev/docs/controls/dropdown
    """

    def __init__(
        self,
        value: Optional[str] = None,
        options: Optional[List[Option]] = None,
        alignment: Optional[Alignment] = None,
        autofocus: Optional[bool] = None,
        hint_content: Optional[Control] = None,
        select_icon: Optional[IconValueOrControl] = None,
        elevation: OptionalNumber = None,
        item_height: OptionalNumber = None,
        max_menu_height: OptionalNumber = None,
        select_icon_size: OptionalNumber = None,
        enable_feedback: Optional[bool] = None,
        padding: Optional[PaddingValue] = None,
        select_icon_enabled_color: Optional[ColorValue] = None,
        select_icon_disabled_color: Optional[ColorValue] = None,
        options_fill_horizontally: Optional[bool] = None,
        disabled_hint_content: Optional[Control] = None,
        on_change: OptionalControlEventCallable = None,
        on_focus: OptionalControlEventCallable = None,
        on_blur: OptionalControlEventCallable = None,
        on_click: OptionalControlEventCallable = None,
        #
        # FormField specific
        #
        text_size: OptionalNumber = None,
        text_style: Optional[TextStyle] = None,
        label: Optional[Union[str, Control]] = None,
        label_style: Optional[TextStyle] = None,
        icon: Optional[IconValueOrControl] = None,
        border: Optional[InputBorder] = None,
        color: Optional[ColorValue] = None,
        bgcolor: Optional[ColorValue] = None,
        border_radius: Optional[BorderRadiusValue] = None,
        border_width: OptionalNumber = None,
        border_color: Optional[ColorValue] = None,
        focused_color: Optional[ColorValue] = None,
        focused_bgcolor: Optional[ColorValue] = None,
        focused_border_width: OptionalNumber = None,
        focused_border_color: Optional[ColorValue] = None,
        content_padding: Optional[PaddingValue] = None,
        dense: Optional[bool] = None,
        filled: Optional[bool] = None,
        fill_color: Optional[ColorValue] = None,
        hint_text: Optional[str] = None,
        hint_style: Optional[TextStyle] = None,
        helper: Optional[Control] = None,
        helper_text: Optional[str] = None,
        helper_style: Optional[TextStyle] = None,
        counter: Optional[Control] = None,
        counter_text: Optional[str] = None,
        counter_style: Optional[TextStyle] = None,
        error: Optional[Control] = None,
        error_text: Optional[str] = None,
        error_style: Optional[TextStyle] = None,
        prefix: Optional[Control] = None,
        prefix_icon: Optional[IconValueOrControl] = None,
        prefix_text: Optional[str] = None,
        prefix_style: Optional[TextStyle] = None,
        suffix: Optional[Control] = None,
        suffix_icon: Optional[IconValueOrControl] = None,
        suffix_text: Optional[str] = None,
        suffix_style: Optional[TextStyle] = None,
        focus_color: Optional[ColorValue] = None,
        align_label_with_hint: Optional[bool] = None,
        hint_fade_duration: Optional[DurationValue] = None,
        hint_max_lines: Optional[int] = None,
        helper_max_lines: Optional[int] = None,
        error_max_lines: Optional[int] = None,
        prefix_icon_size_constraints: Optional[BoxConstraints] = None,
        suffix_icon_size_constraints: Optional[BoxConstraints] = None,
        size_constraints: Optional[BoxConstraints] = None,
        collapsed: Optional[bool] = None,
        #
        # ConstrainedControl
        #
        ref: Optional[Ref] = None,
        key: Optional[str] = None,
        width: OptionalNumber = None,
        height: OptionalNumber = None,
        expand: Union[None, bool, int] = None,
        expand_loose: Optional[bool] = None,
        col: Optional[ResponsiveNumber] = None,
        opacity: OptionalNumber = None,
        rotate: Optional[RotateValue] = None,
        scale: Optional[ScaleValue] = None,
        offset: Optional[OffsetValue] = None,
        aspect_ratio: OptionalNumber = None,
        animate_opacity: Optional[AnimationValue] = None,
        animate_size: Optional[AnimationValue] = None,
        animate_position: Optional[AnimationValue] = None,
        animate_rotation: Optional[AnimationValue] = None,
        animate_scale: Optional[AnimationValue] = None,
        animate_offset: Optional[AnimationValue] = None,
        on_animation_end: OptionalControlEventCallable = None,
        tooltip: Optional[TooltipValue] = None,
        badge: Optional[BadgeValue] = None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
    ):
        FormFieldControl.__init__(
            self,
            ref=ref,
            key=key,
            width=width,
            height=height,
            expand=expand,
            expand_loose=expand_loose,
            col=col,
            opacity=opacity,
            rotate=rotate,
            scale=scale,
            offset=offset,
            aspect_ratio=aspect_ratio,
            animate_opacity=animate_opacity,
            animate_size=animate_size,
            animate_position=animate_position,
            animate_rotation=animate_rotation,
            animate_scale=animate_scale,
            animate_offset=animate_offset,
            on_animation_end=on_animation_end,
            tooltip=tooltip,
            badge=badge,
            visible=visible,
            disabled=disabled,
            data=data,
            #
            # FormField specific
            #
            text_size=text_size,
            text_style=text_style,
            label=label,
            label_style=label_style,
            icon=icon,
            border=border,
            color=color,
            bgcolor=bgcolor,
            border_radius=border_radius,
            border_width=border_width,
            border_color=border_color,
            focused_color=focused_color,
            focused_bgcolor=focused_bgcolor,
            focused_border_width=focused_border_width,
            focused_border_color=focused_border_color,
            content_padding=content_padding,
            dense=dense,
            filled=filled,
            fill_color=fill_color,
            hint_text=hint_text,
            hint_style=hint_style,
            helper_text=helper_text,
            helper=helper,
            helper_style=helper_style,
            counter=counter,
            counter_text=counter_text,
            counter_style=counter_style,
            error=error,
            error_text=error_text,
            error_style=error_style,
            prefix=prefix,
            prefix_icon=prefix_icon,
            prefix_text=prefix_text,
            prefix_style=prefix_style,
            suffix=suffix,
            suffix_icon=suffix_icon,
            suffix_text=suffix_text,
            suffix_style=suffix_style,
            focus_color=focus_color,
            align_label_with_hint=align_label_with_hint,
            hint_fade_duration=hint_fade_duration,
            hint_max_lines=hint_max_lines,
            helper_max_lines=helper_max_lines,
            error_max_lines=error_max_lines,
            prefix_icon_size_constraints=prefix_icon_size_constraints,
            suffix_icon_size_constraints=suffix_icon_size_constraints,
            size_constraints=size_constraints,
            collapsed=collapsed,
        )

        self.value = value
        self.autofocus = autofocus
        self.options = options
        self.alignment = alignment
        self.elevation = elevation
        self.hint_content = hint_content
        self.disabled_hint_content = disabled_hint_content
        self.select_icon = select_icon
        self.padding = padding
        self.enable_feedback = enable_feedback
        self.on_focus = on_focus
        self.on_blur = on_blur
        self.on_change = on_change
        self.item_height = item_height
        self.max_menu_height = max_menu_height
        self.select_icon_size = select_icon_size
        self.select_icon_enabled_color = select_icon_enabled_color
        self.select_icon_disabled_color = select_icon_disabled_color

        self.on_click = on_click
        self.options_fill_horizontally = options_fill_horizontally

        warnings.warn(
            f"DropdownM2 control is deprecated since version 0.27.0 "
            f"and will be removed in version 0.30.0. User Dropdown control instead",
            category=DeprecationWarning,
            stacklevel=2,
        )

    def _get_control_name(self):
        return "dropdownm2"

    def before_update(self):
        super().before_update()
        self._set_attr_json("padding", self.__padding)
        self._set_attr_json("alignment", self.__alignment)
        if (
            (
                self.bgcolor is not None
                or self.fill_color is not None
                or self.focused_bgcolor is not None
            )
        ) and self.filled is None:
            self.filled = True  # required to display any of the above colors

    def _get_children(self):
        children = FormFieldControl._get_children(self) + self.__options
        if isinstance(self.__hint_content, Control):
            self.__hint_content._set_attr_internal("n", "hint")
            children.append(self.__hint_content)
        if isinstance(self.__select_icon, Control):
            self.__select_icon._set_attr_internal("n", "selectIcon")
            children.append(self.__select_icon)
        if isinstance(self.__disabled_hint_content, Control):
            self.__disabled_hint_content._set_attr_internal("n", "disabled_hint")
            children.append(self.__disabled_hint_content)
        return children

    def __contains__(self, item):
        return item in self.__options

    def focus(self):
        self._set_attr_json("focus", str(time.time()))
        self.update()

    # options
    @property
    def options(self):
        return self.__options

    @options.setter
    def options(self, value):
        self.__options = value if value is not None else []

    # select_icon
    @property
    def select_icon(self) -> Optional[IconValueOrControl]:
        return self.__select_icon

    @select_icon.setter
    def select_icon(self, value: Optional[IconValueOrControl]):
        self.__select_icon = value
        if not isinstance(value, Control):
            self._set_enum_attr("selectIcon", value, IconEnums)

    # hint_content
    @property
    def hint_content(self) -> Optional[Control]:
        return self.__hint_content

    @hint_content.setter
    def hint_content(self, value: Optional[Control]):
        self.__hint_content = value

    # disabled_hint_content
    @property
    def disabled_hint_content(self) -> Optional[Control]:
        return self.__disabled_hint_content

    @disabled_hint_content.setter
    def disabled_hint_content(self, value: Optional[Control]):
        self.__disabled_hint_content = value

    # value
    @property
    def value(self) -> Optional[str]:
        return self._get_attr("value")

    @value.setter
    def value(self, value: Optional[str]):
        self._set_attr("value", value)

    # select_icon_enabled_color
    @property
    def select_icon_enabled_color(self) -> Optional[ColorValue]:
        return self.__select_icon_enabled_color

    @select_icon_enabled_color.setter
    def select_icon_enabled_color(self, value: Optional[ColorValue]):
        self.__select_icon_enabled_color = value
        self._set_enum_attr("selectIconEnabledColor", value, ColorEnums)

    # select_icon_disabled_color
    @property
    def select_icon_disabled_color(self) -> Optional[ColorValue]:
        return self.__select_icon_disabled_color

    @select_icon_disabled_color.setter
    def select_icon_disabled_color(self, value: Optional[ColorValue]):
        self.__select_icon_disabled_color = value
        self._set_enum_attr("selectIconDisabledColor", value, ColorEnums)

    # item_height
    @property
    def item_height(self) -> OptionalNumber:
        return self._get_attr("itemHeight", data_type="float")

    @item_height.setter
    def item_height(self, value: OptionalNumber):
        assert (
            value is None or value >= 48.0
        ), "item_height must be greater than or equal to 48.0"
        self._set_attr("itemHeight", value)

    # max_menu_height
    @property
    def max_menu_height(self) -> OptionalNumber:
        return self._get_attr("maxMenuHeight", data_type="float")

    @max_menu_height.setter
    def max_menu_height(self, value: OptionalNumber):
        self._set_attr("maxMenuHeight", value)

    # select_icon_size
    @property
    def select_icon_size(self) -> float:
        return self._get_attr("selectIconSize", data_type="float", def_value=24.0)

    @select_icon_size.setter
    def select_icon_size(self, value: OptionalNumber):
        self._set_attr("selectIconSize", value)

    # padding
    @property
    def padding(self) -> Optional[PaddingValue]:
        return self.__padding

    @padding.setter
    def padding(self, value: Optional[PaddingValue]):
        self.__padding = value

    # autofocus
    @property
    def autofocus(self) -> bool:
        return self._get_attr("autofocus", data_type="bool", def_value=False)

    @autofocus.setter
    def autofocus(self, value: Optional[bool]):
        self._set_attr("autofocus", value)

    # options_fill_horizontally
    @property
    def options_fill_horizontally(self) -> bool:
        return self._get_attr(
            "optionsFillHorizontally", data_type="bool", def_value=False
        )

    @options_fill_horizontally.setter
    def options_fill_horizontally(self, value: Optional[bool]):
        self._set_attr("optionsFillHorizontally", value)

    # enable_feedback
    @property
    def enable_feedback(self) -> bool:
        return self._get_attr("enableFeedback", data_type="bool", def_value=True)

    @enable_feedback.setter
    def enable_feedback(self, value: Optional[bool]):
        self._set_attr("enableFeedback", value)

    # elevation
    @property
    def elevation(self) -> float:
        return self._get_attr("elevation", data_type="float", def_value=8.0)

    @elevation.setter
    def elevation(self, value: OptionalNumber):
        self._set_attr("elevation", value)

    # alignment
    @property
    def alignment(self) -> Optional[Alignment]:
        return self.__alignment

    @alignment.setter
    def alignment(self, value: Optional[Alignment]):
        self.__alignment = value

    # on_change
    @property
    def on_change(self) -> OptionalControlEventCallable:
        return self._get_event_handler("change")

    @on_change.setter
    def on_change(self, handler: OptionalControlEventCallable):
        self._add_event_handler("change", handler)

    # on_focus
    @property
    def on_focus(self) -> OptionalControlEventCallable:
        return self._get_event_handler("focus")

    @on_focus.setter
    def on_focus(self, handler: OptionalControlEventCallable):
        self._add_event_handler("focus", handler)

    # on_blur
    @property
    def on_blur(self) -> OptionalControlEventCallable:
        return self._get_event_handler("blur")

    @on_blur.setter
    def on_blur(self, handler: OptionalControlEventCallable):
        self._add_event_handler("blur", handler)

    # on_click
    @property
    def on_click(self) -> OptionalControlEventCallable:
        return self._get_event_handler("click")

    @on_click.setter
    def on_click(self, handler: OptionalControlEventCallable):
        self._add_event_handler("click", handler)



import json
from enum import Enum
from typing import Any, List, Optional, Union

from flet.core.animation import AnimationValue
from flet.core.badge import BadgeValue
from flet.core.charts.pie_chart_section import PieChartSection
from flet.core.constrained_control import ConstrainedControl
from flet.core.control import OptionalNumber
from flet.core.control_event import ControlEvent
from flet.core.event_handler import EventHandler
from flet.core.ref import Ref
from flet.core.tooltip import TooltipValue
from flet.core.types import (
    ColorEnums,
    ColorValue,
    OffsetValue,
    OptionalControlEventCallable,
    OptionalEventCallable,
    ResponsiveNumber,
    RotateValue,
    ScaleValue,
)


class PieChart(ConstrainedControl):
    def __init__(
        self,
        sections: Optional[List[PieChartSection]] = None,
        center_space_color: Optional[ColorValue] = None,
        center_space_radius: OptionalNumber = None,
        sections_space: OptionalNumber = None,
        start_degree_offset: OptionalNumber = None,
        animate: Optional[AnimationValue] = None,
        on_chart_event: OptionalEventCallable["PieChartEvent"] = None,
        #
        # ConstrainedControl
        #
        ref: Optional[Ref] = None,
        width: OptionalNumber = None,
        height: OptionalNumber = None,
        left: OptionalNumber = None,
        top: OptionalNumber = None,
        right: OptionalNumber = None,
        bottom: OptionalNumber = None,
        expand: Union[None, bool, int] = None,
        expand_loose: Optional[bool] = None,
        col: Optional[ResponsiveNumber] = None,
        opacity: OptionalNumber = None,
        rotate: Optional[RotateValue] = None,
        scale: Optional[ScaleValue] = None,
        offset: Optional[OffsetValue] = None,
        aspect_ratio: OptionalNumber = None,
        animate_opacity: Optional[AnimationValue] = None,
        animate_size: Optional[AnimationValue] = None,
        animate_position: Optional[AnimationValue] = None,
        animate_rotation: Optional[AnimationValue] = None,
        animate_scale: Optional[AnimationValue] = None,
        animate_offset: Optional[AnimationValue] = None,
        on_animation_end: OptionalControlEventCallable = None,
        tooltip: Optional[TooltipValue] = None,
        badge: Optional[BadgeValue] = None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
    ):
        ConstrainedControl.__init__(
            self,
            ref=ref,
            width=width,
            height=height,
            left=left,
            top=top,
            right=right,
            bottom=bottom,
            expand=expand,
            expand_loose=expand_loose,
            col=col,
            opacity=opacity,
            rotate=rotate,
            scale=scale,
            offset=offset,
            aspect_ratio=aspect_ratio,
            animate_opacity=animate_opacity,
            animate_size=animate_size,
            animate_position=animate_position,
            animate_rotation=animate_rotation,
            animate_scale=animate_scale,
            animate_offset=animate_offset,
            on_animation_end=on_animation_end,
            tooltip=tooltip,
            badge=badge,
            visible=visible,
            disabled=disabled,
            data=data,
        )

        self.__on_chart_event = EventHandler(lambda e: PieChartEvent(e))
        self._add_event_handler("chart_event", self.__on_chart_event.get_handler())

        self.sections = sections
        self.center_space_color = center_space_color
        self.center_space_radius = center_space_radius
        self.sections_space = sections_space
        self.start_degree_offset = start_degree_offset
        self.animate = animate
        self.on_chart_event = on_chart_event

    def _get_control_name(self):
        return "piechart"

    def before_update(self):
        super().before_update()
        self._set_attr_json("animate", self.__animate)

    def _get_children(self):
        children = []
        for ds in self.__sections:
            children.append(ds)
        return children

    # sections
    @property
    def sections(self):
        return self.__sections

    @sections.setter
    def sections(self, value):
        self.__sections = value if value is not None else []

    # center_space_color
    @property
    def center_space_color(self) -> Optional[ColorValue]:
        return self.__center_space_color

    @center_space_color.setter
    def center_space_color(self, value: Optional[ColorValue]):
        self.__center_space_color = value
        self._set_enum_attr("centerSpaceColor", value, ColorEnums)

    # center_space_radius
    @property
    def center_space_radius(self) -> OptionalNumber:
        return self._get_attr("centerSpaceRadius", data_type="float")

    @center_space_radius.setter
    def center_space_radius(self, value: OptionalNumber):
        self._set_attr("centerSpaceRadius", value)

    # sections_space
    @property
    def sections_space(self) -> OptionalNumber:
        return self._get_attr("sectionsSpace", data_type="float")

    @sections_space.setter
    def sections_space(self, value: OptionalNumber):
        self._set_attr("sectionsSpace", value)

    # start_degree_offset
    @property
    def start_degree_offset(self) -> OptionalNumber:
        return self._get_attr("startDegreeOffset", data_type="float")

    @start_degree_offset.setter
    def start_degree_offset(self, value: OptionalNumber):
        self._set_attr("startDegreeOffset", value)

    # animate
    @property
    def animate(self) -> AnimationValue:
        return self.__animate

    @animate.setter
    def animate(self, value: AnimationValue):
        self.__animate = value

    # on_chart_event
    @property
    def on_chart_event(self) -> OptionalEventCallable["PieChartEvent"]:
        return self.__on_chart_event.handler

    @on_chart_event.setter
    def on_chart_event(self, handler: OptionalEventCallable["PieChartEvent"]):
        self.__on_chart_event.handler = handler
        self._set_attr("onChartEvent", True if handler is not None else None)


class PieChartEventType(Enum):
    POINTER_ENTER = "pointerEnter"
    POINTER_EXIT = "pointerExit"
    POINTER_HOVER = "pointerHover"
    PAN_CANCEL = "panCancel"
    PAN_DOWN = "panDown"
    PAN_END = "panEnd"
    PAN_START = "panStart"
    PAN_UPDATE = "panUpdate"
    LONG_PRESS_END = "longPressEnd"
    LONG_PRESS_MOVE_UPDATE = "longPressMoveUpdate"
    LONG_PRESS_START = "longPressStart"
    TAP_CANCEL = "tapCancel"
    TAP_DOWN = "tapDown"
    TAP_UP = "tapUp"
    UNDEFINED = "undefined"


class PieChartEvent(ControlEvent):
    def __init__(self, e: ControlEvent):
        super().__init__(e.target, e.name, e.data, e.control, e.page)
        d = json.loads(e.data)
        self.type: PieChartEventType = PieChartEventType(d.get("type"))
        self.section_index: int = d.get("section_index")
        self.local_x: Optional[float] = d.get("lx")
        self.local_y: Optional[float] = d.get("ly")
        # self.radius: float = d["radius"]
        # self.angle: float = d["angle"]



import json
from typing import Any, List, Optional, Set, Union

from flet.core.alignment import Axis
from flet.core.animation import AnimationValue
from flet.core.badge import BadgeValue
from flet.core.buttons import ButtonStyle
from flet.core.constrained_control import ConstrainedControl
from flet.core.control import Control, OptionalNumber
from flet.core.ref import Ref
from flet.core.tooltip import TooltipValue
from flet.core.types import (
    OffsetValue,
    OptionalControlEventCallable,
    PaddingValue,
    ResponsiveNumber,
    RotateValue,
    ScaleValue,
)


class Segment(Control):
    def __init__(
        self,
        value: str,
        icon: Optional[Control] = None,
        label: Optional[Control] = None,
        #
        # Control
        #
        ref: Optional[Ref] = None,
        expand: Union[None, bool, int] = None,
        expand_loose: Optional[bool] = None,
        col: Optional[ResponsiveNumber] = None,
        opacity: OptionalNumber = None,
        tooltip: Optional[str] = None,
        badge: Optional[BadgeValue] = None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
    ):
        Control.__init__(
            self,
            ref=ref,
            expand=expand,
            expand_loose=expand_loose,
            col=col,
            opacity=opacity,
            badge=badge,
            visible=visible,
            disabled=disabled,
            data=data,
        )

        self.value = value
        self.label = label
        self.icon = icon
        self.tooltip = tooltip

    def _get_control_name(self):
        return "segment"

    def _get_children(self):
        children = []
        if self.__label:
            self.__label._set_attr_internal("n", "label")
            children.append(self.__label)
        if self.__icon:
            self.__icon._set_attr_internal("n", "icon")
            children.append(self.__icon)
        return children

    # label
    @property
    def label(self) -> Optional[Control]:
        return self.__label

    @label.setter
    def label(self, value: Optional[Control]):
        self.__label = value

    # icon
    @property
    def icon(self) -> Optional[Control]:
        return self.__icon

    @icon.setter
    def icon(self, value: Optional[Control]):
        self.__icon = value

    # value
    @property
    def value(self) -> str:
        return self._get_attr("value")

    @value.setter
    def value(self, value: str):
        self._set_attr("value", value)

    # tooltip
    @property
    def tooltip(self) -> Optional[str]:
        return self._get_attr("tooltip")

    @tooltip.setter
    def tooltip(self, value: Optional[str]):
        self._set_attr("tooltip", value)


class SegmentedButton(ConstrainedControl):
    """
    A segmented button control.

    -----

    Online docs: https://flet.dev/docs/controls/segmentedbutton
    """

    def __init__(
        self,
        segments: List[Segment],
        style: Optional[ButtonStyle] = None,
        allow_empty_selection: Optional[bool] = None,
        allow_multiple_selection: Optional[bool] = None,
        selected: Optional[Set] = None,
        selected_icon: Optional[Control] = None,
        show_selected_icon: Optional[bool] = None,
        direction: Optional[Axis] = None,
        padding: Optional[PaddingValue] = None,
        on_change: OptionalControlEventCallable = None,
        #
        # ConstrainedControl
        #
        ref: Optional[Ref] = None,
        key: Optional[str] = None,
        width: OptionalNumber = None,
        height: OptionalNumber = None,
        left: OptionalNumber = None,
        top: OptionalNumber = None,
        right: OptionalNumber = None,
        bottom: OptionalNumber = None,
        expand: Union[None, bool, int] = None,
        expand_loose: Optional[bool] = None,
        col: Optional[ResponsiveNumber] = None,
        opacity: OptionalNumber = None,
        rotate: Optional[RotateValue] = None,
        scale: Optional[ScaleValue] = None,
        offset: Optional[OffsetValue] = None,
        aspect_ratio: OptionalNumber = None,
        animate_opacity: Optional[AnimationValue] = None,
        animate_size: Optional[AnimationValue] = None,
        animate_position: Optional[AnimationValue] = None,
        animate_rotation: Optional[AnimationValue] = None,
        animate_scale: Optional[AnimationValue] = None,
        animate_offset: Optional[AnimationValue] = None,
        on_animation_end: OptionalControlEventCallable = None,
        tooltip: Optional[TooltipValue] = None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
    ):
        ConstrainedControl.__init__(
            self,
            ref=ref,
            key=key,
            width=width,
            height=height,
            left=left,
            top=top,
            right=right,
            bottom=bottom,
            expand=expand,
            expand_loose=expand_loose,
            col=col,
            opacity=opacity,
            rotate=rotate,
            scale=scale,
            offset=offset,
            aspect_ratio=aspect_ratio,
            animate_opacity=animate_opacity,
            animate_size=animate_size,
            animate_position=animate_position,
            animate_rotation=animate_rotation,
            animate_scale=animate_scale,
            animate_offset=animate_offset,
            on_animation_end=on_animation_end,
            tooltip=tooltip,
            visible=visible,
            disabled=disabled,
            data=data,
        )

        self.segments = segments
        self.show_selected_icon = show_selected_icon
        self.allow_multiple_selection = allow_multiple_selection
        self.allow_empty_selection = allow_empty_selection
        self.selected_icon = selected_icon
        self.selected = selected
        self.style = style
        self.direction = direction
        self.padding = padding
        self.on_change = on_change

    def _get_control_name(self):
        return "segmentedbutton"

    def before_update(self):
        super().before_update()
        assert any(
            segment.visible for segment in self.__segments
        ), "segments must have at minimum one visible Segment"
        assert (
            len(self.selected) > 0 or self.allow_empty_selection
        ), "allow_empty_selection must be True for selected to be empty"
        assert (
            len(self.selected) < 2 or self.allow_multiple_selection
        ), "allow_multiple_selection must be True for selected to have more than one item"
        style = self.__style or ButtonStyle()
        style.side = self._wrap_attr_dict(style.side)
        style.shape = self._wrap_attr_dict(style.shape)
        style.padding = self._wrap_attr_dict(style.padding)
        self._set_attr_json("style", style)
        self._set_attr_json("padding", self.__padding)

    def _get_children(self):
        for segment in self.segments:
            segment._set_attr_internal("n", "segment")
        children: List[Control] = self.__segments
        if self.__selected_icon:
            self.__selected_icon._set_attr_internal("n", "selectedIcon")
            children.append(self.__selected_icon)
        return children

    def __contains__(self, item):
        return item in self.__segments

    # style
    @property
    def style(self) -> Optional[ButtonStyle]:
        return self.__style

    @style.setter
    def style(self, value: Optional[ButtonStyle]):
        self.__style = value

    # on_change
    @property
    def on_change(self) -> OptionalControlEventCallable:
        return self._get_event_handler("change")

    @on_change.setter
    def on_change(self, handler: OptionalControlEventCallable):
        self._add_event_handler("change", handler)

    # segments
    @property
    def segments(self) -> List[Segment]:
        return self.__segments

    @segments.setter
    def segments(self, value: List[Segment]):
        self.__segments = value

    # padding
    @property
    def padding(self) -> Optional[PaddingValue]:
        return self.__padding

    @padding.setter
    def padding(self, value: Optional[PaddingValue]):
        self.__padding = value

    # allow_empty_selection
    @property
    def allow_empty_selection(self) -> bool:
        return self._get_attr("allowEmptySelection", data_type="bool", def_value=False)

    @allow_empty_selection.setter
    def allow_empty_selection(self, value: Optional[bool]):
        self._set_attr("allowEmptySelection", value)

    # allow_multiple_selection
    @property
    def allow_multiple_selection(self) -> bool:
        return self._get_attr(
            "allowMultipleSelection", data_type="bool", def_value=False
        )

    @allow_multiple_selection.setter
    def allow_multiple_selection(self, value: Optional[bool]):
        self._set_attr("allowMultipleSelection", value)

    # selected
    @property
    def selected(self) -> Optional[Set]:
        s = self._get_attr("selected")
        return set(json.loads(s)) if s else s

    @selected.setter
    def selected(self, value: Optional[Set]):
        self._set_attr(
            "selected",
            (
                json.dumps(list(value), separators=(",", ":"))
                if value is not None
                else None
            ),
        )

    # show_selected_icon
    @property
    def show_selected_icon(self) -> bool:
        return self._get_attr("showSelectedIcon", data_type="bool", def_value=True)

    @show_selected_icon.setter
    def show_selected_icon(self, value: Optional[bool]):
        self._set_attr("showSelectedIcon", value)

    # direction
    @property
    def direction(self) -> Optional[Axis]:
        return self.__direction

    @direction.setter
    def direction(self, value: Optional[Axis]):
        self.__direction = value
        self._set_enum_attr("direction", value, Axis)

    # selected_icon
    @property
    def selected_icon(self) -> Optional[Control]:
        return self.__selected_icon

    @selected_icon.setter
    def selected_icon(self, value: Optional[Control]):
        self.__selected_icon = value



from typing import Any, List, Optional, Union

from flet.core.animation import AnimationValue
from flet.core.border import Border
from flet.core.constrained_control import ConstrainedControl
from flet.core.navigation_bar import NavigationBarDestination
from flet.core.ref import Ref
from flet.core.types import (
    ColorEnums,
    ColorValue,
    OffsetValue,
    OptionalControlEventCallable,
    OptionalNumber,
    ResponsiveNumber,
    RotateValue,
    ScaleValue,
)


class CupertinoNavigationBar(ConstrainedControl):
    """
    An iOS-styled bottom navigation tab bar.

    Navigation bars offer a persistent and convenient way to switch between primary destinations in an app.

    Example:

    ```
    import flet as ft

    def main(page: ft.Page):
        page.title = "CupertinoNavigationBar Example"
        page.navigation_bar = ft.CupertinoNavigationBar(
            bgcolor=ft.colors.AMBER_100,
            inactive_color=ft.colors.GREY,
            active_color=ft.colors.BLACK,
            on_change=lambda e: print("Selected tab:", e.control.selected_index),
            destinations=[
                ft.NavigationBarDestination(icon=ft.icons.EXPLORE, label="Explore"),
                ft.NavigationBarDestination(icon=ft.icons.COMMUTE, label="Commute"),
                ft.NavigationBarDestination(
                    icon=ft.icons.BOOKMARK_BORDER,
                    selected_icon=ft.icons.BOOKMARK,
                    label="Explore",
                ),
            ]
        )
        page.add(ft.SafeArea(ft.Text("Body!")))


    ft.app(target=main)
    ```

    -----

    Online docs: https://flet.dev/docs/controls/cupertinonavigationbar
    """

    def __init__(
        self,
        destinations: Optional[List[NavigationBarDestination]] = None,
        selected_index: Optional[int] = None,
        bgcolor: Optional[ColorValue] = None,
        active_color: Optional[ColorValue] = None,
        inactive_color: Optional[ColorValue] = None,
        border: Optional[Border] = None,
        icon_size: OptionalNumber = None,
        on_change: OptionalControlEventCallable = None,
        #
        # ConstrainedControl
        #
        ref: Optional[Ref] = None,
        width: OptionalNumber = None,
        height: OptionalNumber = None,
        left: OptionalNumber = None,
        top: OptionalNumber = None,
        right: OptionalNumber = None,
        bottom: OptionalNumber = None,
        expand: Union[None, bool, int] = None,
        expand_loose: Optional[bool] = None,
        col: Optional[ResponsiveNumber] = None,
        opacity: OptionalNumber = None,
        rotate: Optional[RotateValue] = None,
        scale: Optional[ScaleValue] = None,
        offset: Optional[OffsetValue] = None,
        aspect_ratio: OptionalNumber = None,
        animate_opacity: Optional[AnimationValue] = None,
        animate_size: Optional[AnimationValue] = None,
        animate_position: Optional[AnimationValue] = None,
        animate_rotation: Optional[AnimationValue] = None,
        animate_scale: Optional[AnimationValue] = None,
        animate_offset: Optional[AnimationValue] = None,
        on_animation_end: OptionalControlEventCallable = None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
    ):
        ConstrainedControl.__init__(
            self,
            ref=ref,
            width=width,
            height=height,
            left=left,
            top=top,
            right=right,
            bottom=bottom,
            expand=expand,
            expand_loose=expand_loose,
            col=col,
            opacity=opacity,
            rotate=rotate,
            scale=scale,
            offset=offset,
            aspect_ratio=aspect_ratio,
            animate_opacity=animate_opacity,
            animate_size=animate_size,
            animate_position=animate_position,
            animate_rotation=animate_rotation,
            animate_scale=animate_scale,
            animate_offset=animate_offset,
            on_animation_end=on_animation_end,
            visible=visible,
            disabled=disabled,
            data=data,
        )

        self.destinations = destinations
        self.selected_index = selected_index
        self.bgcolor = bgcolor
        self.active_color = active_color
        self.inactive_color = inactive_color
        self.border = border
        self.icon_size = icon_size
        self.on_change = on_change

    def _get_control_name(self):
        return "cupertinonavigationbar"

    def before_update(self) -> None:
        super().before_update()
        self._set_attr_json("border", self.__border)

    def _get_children(self):
        return self.__destinations

    # destinations
    @property
    def destinations(self) -> Optional[List[NavigationBarDestination]]:
        return self.__destinations

    @destinations.setter
    def destinations(self, value: Optional[List[NavigationBarDestination]]):
        self.__destinations = value if value else []

    # border
    @property
    def border(self) -> Optional[Border]:
        return self.__border

    @border.setter
    def border(self, value: Optional[Border]):
        self.__border = value

    # selected_index
    @property
    def selected_index(self) -> int:
        return self._get_attr("selectedIndex", data_type="int", def_value=0)

    @selected_index.setter
    def selected_index(self, value: Optional[int]):
        self._set_attr("selectedIndex", value)

    # bgcolor
    @property
    def bgcolor(self) -> Optional[ColorValue]:
        return self.__bgcolor

    @bgcolor.setter
    def bgcolor(self, value: Optional[ColorValue]):
        self.__bgcolor = value
        self._set_enum_attr("bgcolor", value, ColorEnums)

    # active_color
    @property
    def active_color(self) -> Optional[ColorValue]:
        return self.__active_color

    @active_color.setter
    def active_color(self, value: Optional[ColorValue]):
        self.__active_color = value
        self._set_enum_attr("activeColor", value, ColorEnums)

    # inactive_color
    @property
    def inactive_color(self) -> Optional[ColorValue]:
        return self.__inactive_color

    @inactive_color.setter
    def inactive_color(self, value: Optional[ColorValue]):
        self.__inactive_color = value
        self._set_enum_attr("inactiveColor", value, ColorEnums)

    # icon_size
    @property
    def icon_size(self) -> OptionalNumber:
        return self._get_attr("iconSize")

    @icon_size.setter
    def icon_size(self, value: OptionalNumber):
        self._set_attr("iconSize", value)

    # on_change
    @property
    def on_change(self) -> OptionalControlEventCallable:
        return self._get_event_handler("change")

    @on_change.setter
    def on_change(self, handler: OptionalControlEventCallable):
        self._add_event_handler("change", handler)



from enum import Enum
from typing import Any, Optional

from flet.core.buttons import OutlinedBorder
from flet.core.control import Control, OptionalNumber
from flet.core.ref import Ref
from flet.core.types import (
    ClipBehavior,
    ColorEnums,
    ColorValue,
    MarginValue,
    OptionalControlEventCallable,
    PaddingValue,
)


class SnackBarBehavior(Enum):
    FIXED = "fixed"
    FLOATING = "floating"


class DismissDirection(Enum):
    NONE = "none"
    VERTICAL = "vertical"
    HORIZONTAL = "horizontal"
    END_TO_START = "endToStart"
    START_TO_END = "startToEnd"
    UP = "up"
    DOWN = "down"


class SnackBar(Control):
    """
    A lightweight message with an optional action which briefly displays at the bottom of the screen.

    Example:
    ```
    import flet as ft

    class Data:
        def __init__(self) -> None:
            self.counter = 0

    d = Data()

    def main(page):

        page.snack_bar = ft.SnackBar(
            content=ft.Text("Hello, world!"),
            action="Alright!",
        )
        page.snack_bar.open = True

        def on_click(e):
            page.snack_bar = ft.SnackBar(ft.Text(f"Hello {d.counter}"))
            page.snack_bar.open = True
            d.counter += 1
            page.update()

        page.add(ft.ElevatedButton("Open SnackBar", on_click=on_click))

    ft.app(target=main)
    ```

    -----

    Online docs: https://flet.dev/docs/controls/snackbar
    """

    def __init__(
        self,
        content: Control,
        open: bool = False,
        behavior: Optional[SnackBarBehavior] = None,
        dismiss_direction: Optional[DismissDirection] = None,
        show_close_icon: Optional[bool] = False,
        action: Optional[str] = None,
        action_color: Optional[ColorValue] = None,
        close_icon_color: Optional[ColorValue] = None,
        bgcolor: Optional[ColorValue] = None,
        duration: Optional[int] = None,
        margin: Optional[MarginValue] = None,
        padding: Optional[PaddingValue] = None,
        width: OptionalNumber = None,
        elevation: OptionalNumber = None,
        shape: Optional[OutlinedBorder] = None,
        clip_behavior: Optional[ClipBehavior] = None,
        action_overflow_threshold: OptionalNumber = None,
        on_action: OptionalControlEventCallable = None,
        on_visible: OptionalControlEventCallable = None,
        #
        # Control
        #
        ref: Optional[Ref] = None,
        disabled: Optional[bool] = None,
        visible: Optional[bool] = None,
        data: Any = None,
    ):
        Control.__init__(
            self,
            ref=ref,
            disabled=disabled,
            visible=visible,
            data=data,
        )

        self.open = open
        self.behavior = behavior
        self.dismiss_direction = dismiss_direction
        self.show_close_icon = show_close_icon
        self.close_icon_color = close_icon_color
        self.margin = margin
        self.padding = padding
        self.width = width
        self.content = content
        self.action = action
        self.action_color = action_color
        self.bgcolor = bgcolor
        self.duration = duration
        self.elevation = elevation
        self.on_action = on_action
        self.on_visible = on_visible
        self.shape = shape
        self.clip_behavior = clip_behavior
        self.action_overflow_threshold = action_overflow_threshold

    def _get_control_name(self):
        return "snackbar"

    def _get_children(self):
        self.__content._set_attr_internal("n", "content")
        return [self.__content]

    def before_update(self):
        super().before_update()
        self._set_attr_json("shape", self.__shape)
        self._set_attr_json("padding", self.__padding)
        self._set_attr_json("margin", self.__margin)

    # open
    @property
    def open(self) -> bool:
        return self._get_attr("open", data_type="bool", def_value=False)

    @open.setter
    def open(self, value: Optional[bool]):
        self._set_attr("open", value)

    # show_close_icon
    @property
    def show_close_icon(self) -> bool:
        return self._get_attr("showCloseIcon", data_type="bool", def_value=False)

    @show_close_icon.setter
    def show_close_icon(self, value: Optional[bool]):
        self._set_attr("showCloseIcon", value)

    # content
    @property
    def content(self) -> Control:
        return self.__content

    @content.setter
    def content(self, value: Control):
        self.__content = value

    # action
    @property
    def action(self) -> Optional[str]:
        return self._get_attr("action")

    @action.setter
    def action(self, value: Optional[str]):
        self._set_attr("action", value)

    # action_color
    @property
    def action_color(self) -> Optional[ColorValue]:
        return self.__action_color

    @action_color.setter
    def action_color(self, value: Optional[ColorValue]):
        self.__action_color = value
        self._set_enum_attr("actionColor", value, ColorEnums)

    # bgcolor
    @property
    def bgcolor(self) -> Optional[ColorValue]:
        return self.__bgcolor

    @bgcolor.setter
    def bgcolor(self, value: Optional[ColorValue]):
        self.__bgcolor = value
        self._set_enum_attr("bgColor", value, ColorEnums)

    # close_icon_color
    @property
    def close_icon_color(self) -> Optional[ColorValue]:
        return self.__close_icon_color

    @close_icon_color.setter
    def close_icon_color(self, value: Optional[ColorValue]):
        self.__close_icon_color = value
        self._set_enum_attr("closeIconColor", value, ColorEnums)

    # duration
    @property
    def duration(self) -> Optional[int]:
        return self._get_attr("duration", data_type="int")

    @duration.setter
    def duration(self, value: Optional[int]):
        self._set_attr("duration", value)

    # action_overflow_threshold
    @property
    def action_overflow_threshold(self) -> float:
        return self._get_attr(
            "actionOverflowThreshold", data_type="float", def_value=0.25
        )

    @action_overflow_threshold.setter
    def action_overflow_threshold(self, value: OptionalNumber):
        assert (
            value is None or 0 <= value <= 1
        ), "action_overflow_threshold must be between 0 and 1 inclusive"
        self._set_attr("actionOverflowThreshold", value)

    # behavior
    @property
    def behavior(self) -> Optional[SnackBarBehavior]:
        return self.__behavior

    @behavior.setter
    def behavior(self, value: Optional[SnackBarBehavior]):
        self.__behavior = value
        self._set_enum_attr("behavior", value, SnackBarBehavior)

    # dismissDirection
    @property
    def dismiss_direction(self) -> Optional[DismissDirection]:
        return self.__dismiss_direction

    @dismiss_direction.setter
    def dismiss_direction(self, value: Optional[DismissDirection]):
        self.__dismiss_direction = value
        self._set_enum_attr("dismissDirection", value, DismissDirection)

    # padding
    @property
    def padding(self) -> Optional[PaddingValue]:
        return self.__padding

    @padding.setter
    def padding(self, value: Optional[PaddingValue]):
        self.__padding = value

    # margin
    @property
    def margin(self) -> Optional[MarginValue]:
        return self.__margin

    @margin.setter
    def margin(self, value: Optional[MarginValue]):
        self.__margin = value

    # width
    @property
    def width(self) -> OptionalNumber:
        return self._get_attr("width", data_type="float")

    @width.setter
    def width(self, value: OptionalNumber):
        self._set_attr("width", value)

    # elevation
    @property
    def elevation(self) -> OptionalNumber:
        return self._get_attr("elevation", data_type="float")

    @elevation.setter
    def elevation(self, value: OptionalNumber):
        assert value is None or value >= 0, "elevation cannot be negative"
        self._set_attr("elevation", value)

    # clip_behavior
    @property
    def clip_behavior(self) -> Optional[ClipBehavior]:
        return self.__clip_behavior

    @clip_behavior.setter
    def clip_behavior(self, value: Optional[ClipBehavior]):
        self.__clip_behavior = value
        self._set_enum_attr("clipBehavior", value, ClipBehavior)

    # shape
    @property
    def shape(self) -> Optional[OutlinedBorder]:
        return self.__shape

    @shape.setter
    def shape(self, value: Optional[OutlinedBorder]):
        self.__shape = value

    # on_action
    @property
    def on_action(self) -> OptionalControlEventCallable:
        return self._get_event_handler("action")

    @on_action.setter
    def on_action(self, handler: OptionalControlEventCallable):
        self._add_event_handler("action", handler)

    # on_visible
    @property
    def on_visible(self) -> OptionalControlEventCallable:
        return self._get_event_handler("visible")

    @on_visible.setter
    def on_visible(self, handler: OptionalControlEventCallable):
        self._add_event_handler("visible", handler)



from enum import Enum
from typing import Any, Optional, Sequence, Union

from flet.core.adaptive_control import AdaptiveControl
from flet.core.alignment import Alignment
from flet.core.animation import AnimationValue
from flet.core.badge import BadgeValue
from flet.core.buttons import OutlinedBorder
from flet.core.constrained_control import ConstrainedControl
from flet.core.control import Control, OptionalNumber
from flet.core.ref import Ref
from flet.core.tooltip import TooltipValue
from flet.core.types import (
    ClipBehavior,
    ColorEnums,
    ColorValue,
    CrossAxisAlignment,
    OffsetValue,
    OptionalControlEventCallable,
    PaddingValue,
    ResponsiveNumber,
    RotateValue,
    ScaleValue,
    VisualDensity,
)


class TileAffinity(Enum):
    LEADING = "leading"
    TRAILING = "trailing"
    PLATFORM = "platform"


class ExpansionTile(ConstrainedControl, AdaptiveControl):
    """
    A single-line ListTile with an expansion arrow icon that expands or collapses the tile to reveal or hide its controls.

    -----

    Online docs: https://flet.dev/docs/controls/expansiontile
    """

    def __init__(
        self,
        title: Control,
        controls: Optional[Sequence[Control]] = None,
        subtitle: Optional[Control] = None,
        leading: Optional[Control] = None,
        trailing: Optional[Control] = None,
        controls_padding: Optional[PaddingValue] = None,
        tile_padding: Optional[PaddingValue] = None,
        affinity: Optional[TileAffinity] = None,
        expanded_alignment: Optional[Alignment] = None,
        expanded_cross_axis_alignment: CrossAxisAlignment = CrossAxisAlignment.CENTER,
        clip_behavior: Optional[ClipBehavior] = None,
        initially_expanded: Optional[bool] = None,
        maintain_state: Optional[bool] = None,
        text_color: Optional[ColorValue] = None,
        icon_color: Optional[ColorValue] = None,
        shape: Optional[OutlinedBorder] = None,
        bgcolor: Optional[ColorValue] = None,
        collapsed_bgcolor: Optional[ColorValue] = None,
        collapsed_icon_color: Optional[ColorValue] = None,
        collapsed_text_color: Optional[ColorValue] = None,
        collapsed_shape: Optional[OutlinedBorder] = None,
        dense: Optional[bool] = None,
        enable_feedback: Optional[bool] = None,
        show_trailing_icon: Optional[bool] = None,
        min_tile_height: OptionalNumber = None,
        visual_density: Optional[VisualDensity] = None,
        on_change: OptionalControlEventCallable = None,
        #
        # ConstrainedControl
        #
        ref: Optional[Ref] = None,
        key: Optional[str] = None,
        width: OptionalNumber = None,
        height: OptionalNumber = None,
        left: OptionalNumber = None,
        top: OptionalNumber = None,
        right: OptionalNumber = None,
        bottom: OptionalNumber = None,
        expand: Union[None, bool, int] = None,
        expand_loose: Optional[bool] = None,
        col: Optional[ResponsiveNumber] = None,
        opacity: OptionalNumber = None,
        rotate: Optional[RotateValue] = None,
        scale: Optional[ScaleValue] = None,
        offset: Optional[OffsetValue] = None,
        aspect_ratio: OptionalNumber = None,
        animate_opacity: Optional[AnimationValue] = None,
        animate_size: Optional[AnimationValue] = None,
        animate_position: Optional[AnimationValue] = None,
        animate_rotation: Optional[AnimationValue] = None,
        animate_scale: Optional[AnimationValue] = None,
        animate_offset: Optional[AnimationValue] = None,
        on_animation_end: OptionalControlEventCallable = None,
        tooltip: Optional[TooltipValue] = None,
        badge: Optional[BadgeValue] = None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
        #
        # AdaptiveControl
        #
        adaptive: Optional[bool] = None,
    ):
        ConstrainedControl.__init__(
            self,
            ref=ref,
            key=key,
            width=width,
            height=height,
            left=left,
            top=top,
            right=right,
            bottom=bottom,
            expand=expand,
            expand_loose=expand_loose,
            col=col,
            opacity=opacity,
            rotate=rotate,
            scale=scale,
            offset=offset,
            aspect_ratio=aspect_ratio,
            animate_opacity=animate_opacity,
            animate_size=animate_size,
            animate_position=animate_position,
            animate_rotation=animate_rotation,
            animate_scale=animate_scale,
            animate_offset=animate_offset,
            on_animation_end=on_animation_end,
            tooltip=tooltip,
            badge=badge,
            visible=visible,
            disabled=disabled,
            data=data,
        )

        AdaptiveControl.__init__(self, adaptive=adaptive)

        self.controls = controls
        self.controls_padding = controls_padding
        self.expanded_alignment = expanded_alignment
        self.expanded_cross_axis_alignment = expanded_cross_axis_alignment
        self.tile_padding = tile_padding
        self.leading = leading
        self.title = title
        self.subtitle = subtitle
        self.trailing = trailing
        self.affinity = affinity
        self.clip_behavior = clip_behavior
        self.maintain_state = maintain_state
        self.initially_expanded = initially_expanded
        self.shape = shape
        self.text_color = text_color
        self.icon_color = icon_color
        self.bgcolor = bgcolor
        self.collapsed_bgcolor = collapsed_bgcolor
        self.collapsed_icon_color = collapsed_icon_color
        self.collapsed_text_color = collapsed_text_color
        self.collapsed_shape = collapsed_shape
        self.on_change = on_change
        self.dense = dense
        self.enable_feedback = enable_feedback
        self.visual_density = visual_density
        self.show_trailing_icon = show_trailing_icon
        self.min_tile_height = min_tile_height

    def _get_control_name(self):
        return "expansiontile"

    def before_update(self):
        super().before_update()
        assert self.__title.visible, "title must be visible"
        self._set_attr_json("expandedAlignment", self.__expanded_alignment)
        self._set_attr_json("controlsPadding", self.__controls_padding)
        self._set_attr_json("tilePadding", self.__tile_padding)
        self._set_attr_json("shape", self.__shape)
        self._set_attr_json("collapsedShape", self.__collapsed_shape)

    def _get_children(self):
        self.__title._set_attr_internal("n", "title")
        children = [self.__title]
        for c in self.__controls:
            c._set_attr_internal("n", "controls")
            children.append(c)
        if self.__leading:
            self.__leading._set_attr_internal("n", "leading")
            children.append(self.__leading)
        if self.__subtitle:
            self.__subtitle._set_attr_internal("n", "subtitle")
            children.append(self.__subtitle)
        if self.__trailing:
            self.__trailing._set_attr_internal("n", "trailing")
            children.append(self.__trailing)
        return children

    # controls
    @property
    def controls(self):
        return self.__controls

    @controls.setter
    def controls(self, value: Optional[Sequence[Control]]):
        self.__controls = list(value) if value is not None else []

    # controls_padding
    @property
    def controls_padding(self) -> Optional[PaddingValue]:
        return self.__controls_padding

    @controls_padding.setter
    def controls_padding(self, value: Optional[PaddingValue]):
        self.__controls_padding = value

    # tile_padding
    @property
    def tile_padding(self) -> Optional[PaddingValue]:
        return self.__tile_padding

    @tile_padding.setter
    def tile_padding(self, value: Optional[PaddingValue]):
        self.__tile_padding = value

    # expanded_alignment
    @property
    def expanded_alignment(self) -> Optional[Alignment]:
        return self.__expanded_alignment

    @expanded_alignment.setter
    def expanded_alignment(self, value: Optional[Alignment]):
        self.__expanded_alignment = value

    # expanded_cross_axis_alignment
    @property
    def expanded_cross_axis_alignment(self) -> Optional[CrossAxisAlignment]:
        return self.__expanded_cross_axis_alignment

    @expanded_cross_axis_alignment.setter
    def expanded_cross_axis_alignment(self, value: Optional[CrossAxisAlignment]):
        self.__expanded_cross_axis_alignment = value
        self._set_enum_attr("crossAxisAlignment", value, CrossAxisAlignment)

    # affinity
    @property
    def affinity(self) -> Optional[TileAffinity]:
        return self.__affinity

    @affinity.setter
    def affinity(self, value: Optional[TileAffinity]):
        self.__affinity = value
        self._set_enum_attr("affinity", value, TileAffinity)

    # leading
    @property
    def leading(self) -> Optional[Control]:
        return self.__leading

    @leading.setter
    def leading(self, value: Optional[Control]):
        self.__leading = value

    # title
    @property
    def title(self) -> Optional[Control]:
        return self.__title

    @title.setter
    def title(self, value: Optional[Control]):
        self.__title = value

    # subtitle
    @property
    def subtitle(self) -> Optional[Control]:
        return self.__subtitle

    @subtitle.setter
    def subtitle(self, value: Optional[Control]):
        self.__subtitle = value

    # trailing
    @property
    def trailing(self) -> Optional[Control]:
        return self.__trailing

    @trailing.setter
    def trailing(self, value: Optional[Control]):
        self.__trailing = value

    # dense
    @property
    def dense(self) -> bool:
        return self._get_attr("dense", data_type="bool")

    @dense.setter
    def dense(self, value: Optional[bool]):
        self._set_attr("dense", value)

    # enable_feedback
    @property
    def enable_feedback(self) -> bool:
        return self._get_attr("enableFeedback", data_type="bool", def_value=True)

    @enable_feedback.setter
    def enable_feedback(self, value: Optional[bool]):
        self._set_attr("enableFeedback", value)

    # clip_behavior
    @property
    def clip_behavior(self) -> Optional[ClipBehavior]:
        return self.__clip_behavior

    @clip_behavior.setter
    def clip_behavior(self, value: Optional[ClipBehavior]):
        self.__clip_behavior = value
        self._set_enum_attr("clipBehavior", value, ClipBehavior)

    # visual_density
    @property
    def visual_density(self) -> Optional[VisualDensity]:
        return self.__visual_density

    @visual_density.setter
    def visual_density(self, value: Optional[VisualDensity]):
        self.__visual_density = value
        self._set_enum_attr("visualDensity", value, VisualDensity)

    # maintain_state
    @property
    def maintain_state(self) -> bool:
        return self._get_attr("maintainState", data_type="bool", def_value=False)

    @maintain_state.setter
    def maintain_state(self, value: Optional[bool]):
        self._set_attr("maintainState", value)

    # initially_expanded
    @property
    def initially_expanded(self) -> bool:
        return self._get_attr("initiallyExpanded", data_type="bool", def_value=False)

    @initially_expanded.setter
    def initially_expanded(self, value: Optional[bool]):
        self._set_attr("initiallyExpanded", value)

    # shape
    @property
    def shape(self) -> Optional[OutlinedBorder]:
        return self.__shape

    @shape.setter
    def shape(self, value: Optional[OutlinedBorder]):
        self.__shape = value

    # text_color
    @property
    def text_color(self) -> Optional[ColorValue]:
        return self.__text_color

    @text_color.setter
    def text_color(self, value: Optional[ColorValue]):
        self.__text_color = value
        self._set_enum_attr("textColor", value, ColorEnums)

    # icon_color
    @property
    def icon_color(self) -> Optional[ColorValue]:
        return self.__icon_color

    @icon_color.setter
    def icon_color(self, value: Optional[ColorValue]):
        self.__icon_color = value
        self._set_enum_attr("iconColor", value, ColorEnums)

    # bgcolor
    @property
    def bgcolor(self) -> Optional[ColorValue]:
        return self.__bgcolor

    @bgcolor.setter
    def bgcolor(self, value: Optional[ColorValue]):
        self.__bgcolor = value
        self._set_enum_attr("bgColor", value, ColorEnums)

    # collapsed_bgcolor
    @property
    def collapsed_bgcolor(self) -> Optional[str]:
        return self.__collapsed_bgcolor

    @collapsed_bgcolor.setter
    def collapsed_bgcolor(self, value: Optional[str]):
        self.__collapsed_bgcolor = value
        self._set_enum_attr("collapsedBgColor", value, ColorEnums)

    # collapsed_icon_color
    @property
    def collapsed_icon_color(self) -> Optional[ColorValue]:
        return self.__collapsed_icon_color

    @collapsed_icon_color.setter
    def collapsed_icon_color(self, value: Optional[ColorValue]):
        self.__collapsed_icon_color = value
        self._set_enum_attr("collapsedIconColor", value, ColorEnums)

    # collapsed_text_color
    @property
    def collapsed_text_color(self) -> Optional[ColorValue]:
        return self.__collapsed_text_color

    @collapsed_text_color.setter
    def collapsed_text_color(self, value: Optional[ColorValue]):
        self.__collapsed_text_color = value
        self._set_enum_attr("collapsedTextColor", value, ColorEnums)

    # collapsed_shape
    @property
    def collapsed_shape(self) -> Optional[OutlinedBorder]:
        return self.__collapsed_shape

    @collapsed_shape.setter
    def collapsed_shape(self, value: Optional[OutlinedBorder]):
        self.__collapsed_shape = value

    # show_trailing_icon
    @property
    def show_trailing_icon(self) -> bool:
        return self._get_attr("showTrailingIcon", data_type="bool", def_value=True)

    @show_trailing_icon.setter
    def show_trailing_icon(self, value: Optional[bool]):
        self._set_attr("showTrailingIcon", value)

    # min_tile_height
    @property
    def min_tile_height(self) -> OptionalNumber:
        return self._get_attr("minTileHeight")

    @min_tile_height.setter
    def min_tile_height(self, value: OptionalNumber):
        self._set_attr("minTileHeight", value)

    # on_change
    @property
    def on_change(self):
        return self._get_event_handler("change")

    @on_change.setter
    def on_change(self, handler: OptionalControlEventCallable):
        self._add_event_handler("change", handler)
        self._set_attr("onChange", True if handler is not None else None)



from __future__ import annotations

import asyncio
import json
import logging
import threading
import time
import uuid
from concurrent.futures import CancelledError, Future, ThreadPoolExecutor
from contextvars import ContextVar
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from functools import partial
from typing import (
    Any,
    Awaitable,
    Callable,
    Coroutine,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)
from urllib.parse import urlparse

import flet.core
from flet.core.adaptive_control import AdaptiveControl
from flet.core.alert_dialog import AlertDialog
from flet.core.alignment import Alignment
from flet.core.animation import AnimationCurve
from flet.core.app_bar import AppBar
from flet.core.banner import Banner
from flet.core.bottom_app_bar import BottomAppBar
from flet.core.bottom_sheet import BottomSheet
from flet.core.box import BoxDecoration
from flet.core.client_storage import ClientStorage
from flet.core.connection import Connection
from flet.core.control import Control
from flet.core.control_event import ControlEvent
from flet.core.cupertino_alert_dialog import CupertinoAlertDialog
from flet.core.cupertino_app_bar import CupertinoAppBar
from flet.core.cupertino_bottom_sheet import CupertinoBottomSheet
from flet.core.cupertino_navigation_bar import CupertinoNavigationBar
from flet.core.event import Event
from flet.core.event_handler import EventHandler
from flet.core.floating_action_button import FloatingActionButton
from flet.core.locks import NopeLock
from flet.core.navigation_bar import NavigationBar
from flet.core.navigation_drawer import NavigationDrawer, NavigationDrawerPosition
from flet.core.padding import Padding
from flet.core.protocol import Command
from flet.core.pubsub.pubsub_client import PubSubClient
from flet.core.querystring import QueryString
from flet.core.scrollable_control import OnScrollEvent
from flet.core.session_storage import SessionStorage
from flet.core.snack_bar import SnackBar
from flet.core.theme import Theme
from flet.core.types import (
    AppLifecycleState,
    Brightness,
    ColorEnums,
    ColorValue,
    CrossAxisAlignment,
    FloatingActionButtonLocation,
    LocaleConfiguration,
    MainAxisAlignment,
    OffsetValue,
    OptionalControlEventCallable,
    OptionalEventCallable,
    OptionalNumber,
    PaddingValue,
    PagePlatform,
    ScrollMode,
    ThemeMode,
    WindowEventType,
    Wrapper,
)
from flet.core.view import View
from flet.utils import classproperty, is_pyodide

try:
    from typing import ParamSpec
except ImportError:
    from typing_extensions import ParamSpec


logger = logging.getLogger(flet.__name__)
try:
    from typing import ParamSpec
except ImportError:
    from typing_extensions import ParamSpec

_session_page = ContextVar("flet_session_page", default=None)


class context:
    @classproperty
    def page(cls) -> "Page":
        return _session_page.get()


try:
    from flet.auth.authorization import Authorization
    from flet.auth.oauth_provider import OAuthProvider
except ImportError:

    class OAuthProvider: ...

    class Authorization:
        def __init__(
            self,
            provider: OAuthProvider,
            fetch_user: bool,
            fetch_groups: bool,
            scope: Optional[List[str]] = None,
        ): ...


AT = TypeVar("AT", bound=Authorization)

InputT = ParamSpec("InputT")
RetT = TypeVar("RetT")


class PageDisconnectedException(Exception):
    def __init__(self, message):
        super().__init__(message)


class BrowserContextMenu:
    def __init__(self, page: "Page"):
        self.page = page
        self.__disabled = False

    def enable(self, wait_timeout: Optional[float] = 10):
        self.page._invoke_method("enableBrowserContextMenu", wait_timeout=wait_timeout)
        self.__disabled = False

    def disable(self, wait_timeout: Optional[float] = 10):
        self.page._invoke_method("disableBrowserContextMenu", wait_timeout=wait_timeout)
        self.__disabled = True

    @property
    def disabled(self) -> bool:
        return self.__disabled


class Window:
    def __init__(self, page: "Page"):
        self.page = page
        self.__alignment = None
        self.__bgcolor = None
        self.__on_event = EventHandler(lambda e: WindowEvent(e))
        self.page._add_event_handler(
            "window_event",
            self.__on_event.get_handler(),
        )

    # bgcolor
    @property
    def bgcolor(self) -> Optional[ColorValue]:
        return self.__bgcolor

    @bgcolor.setter
    def bgcolor(self, value: Optional[ColorValue]):
        self.__bgcolor = value
        self.page._set_enum_attr("windowBgcolor", value, ColorEnums)

    # width
    @property
    def width(self) -> OptionalNumber:
        w = self.page._get_attr("windowWidth")
        return float(w) if w else 0

    @width.setter
    def width(self, value: OptionalNumber):
        self.page._set_attr("windowWidth", value)

    # height
    @property
    def height(self) -> OptionalNumber:
        h = self.page._get_attr("windowHeight")
        return float(h) if h else 0

    @height.setter
    def height(self, value: OptionalNumber):
        self.page._set_attr("windowHeight", value)

    # top
    @property
    def top(self) -> OptionalNumber:
        w = self.page._get_attr("windowTop")
        return float(w) if w else 0

    @top.setter
    def top(self, value: OptionalNumber):
        self.page._set_attr("windowTop", value)

    # left
    @property
    def left(self) -> OptionalNumber:
        h = self.page._get_attr("windowLeft")
        return float(h) if h else 0

    @left.setter
    def left(self, value: OptionalNumber):
        self.page._set_attr("windowLeft", value)

    # max_width
    @property
    def max_width(self) -> OptionalNumber:
        return self.page._get_attr("windowMaxWidth")

    @max_width.setter
    def max_width(self, value: OptionalNumber):
        self.page._set_attr("windowMaxWidth", value)

    # max_height
    @property
    def max_height(self) -> OptionalNumber:
        return self.page._get_attr("windowMaxHeight")

    @max_height.setter
    def max_height(self, value: OptionalNumber):
        self.page._set_attr("windowMaxHeight", value)

    # min_width
    @property
    def min_width(self) -> OptionalNumber:
        return self.page._get_attr("windowMinWidth")

    @min_width.setter
    def min_width(self, value: OptionalNumber):
        self.page._set_attr("windowMinWidth", value)

    # min_height
    @property
    def min_height(self) -> OptionalNumber:
        return self.page._get_attr("windowMinHeight")

    @min_height.setter
    def min_height(self, value: OptionalNumber):
        self.page._set_attr("windowMinHeight", value)

    # opacity
    @property
    def opacity(self) -> float:
        return self.page._get_attr("windowOpacity", data_type="float", def_value=1.0)

    @opacity.setter
    def opacity(self, value: OptionalNumber):
        self.page._set_attr("windowOpacity", value)

    # maximized
    @property
    def maximized(self) -> bool:
        return self.page._get_attr("windowMaximized", data_type="bool", def_value=False)

    @maximized.setter
    def maximized(self, value: Optional[bool]):
        self.page._set_attr("windowMaximized", value)

    # minimized
    @property
    def minimized(self) -> bool:
        return self.page._get_attr("windowMinimized", data_type="bool", def_value=False)

    @minimized.setter
    def minimized(self, value: Optional[bool]):
        self.page._set_attr("windowMinimized", value)

    # minimizable
    @property
    def minimizable(self) -> bool:
        return self.page._get_attr(
            "windowMinimizable", data_type="bool", def_value=True
        )

    @minimizable.setter
    def minimizable(self, value: Optional[bool]):
        self.page._set_attr("windowMinimizable", value)

    # maximizable
    @property
    def maximizable(self) -> bool:
        return self.page._get_attr(
            "windowMaximizable", data_type="bool", def_value=True
        )

    @maximizable.setter
    def maximizable(self, value: Optional[bool]):
        self.page._set_attr("windowMaximizable", value)

    # resizable
    @property
    def resizable(self) -> bool:
        return self.page._get_attr("windowResizable", data_type="bool", def_value=True)

    @resizable.setter
    def resizable(self, value: Optional[bool]):
        self.page._set_attr("windowResizable", value)

    # movable
    @property
    def movable(self) -> bool:
        return self.page._get_attr("windowMovable", data_type="bool", def_value=True)

    @movable.setter
    def movable(self, value: Optional[bool]):
        self.page._set_attr("windowMovable", value)

    # full_screen
    @property
    def full_screen(self) -> bool:
        return self.page._get_attr(
            "windowFullScreen", data_type="bool", def_value=False
        )

    @full_screen.setter
    def full_screen(self, value: Optional[bool]):
        self.page._set_attr("windowFullScreen", value)

    # always_on_top
    @property
    def always_on_top(self) -> bool:
        return self.page._get_attr(
            "windowAlwaysOnTop", data_type="bool", def_value=False
        )

    @always_on_top.setter
    def always_on_top(self, value: Optional[bool]):
        self.page._set_attr("windowAlwaysOnTop", value)

    # prevent_close
    @property
    def prevent_close(self) -> bool:
        return self.page._get_attr(
            "windowPreventClose", data_type="bool", def_value=False
        )

    @prevent_close.setter
    def prevent_close(self, value: Optional[bool]):
        self.page._set_attr("windowPreventClose", value)

    # title_bar_hidden
    @property
    def title_bar_hidden(self) -> bool:
        return self.page._get_attr(
            "windowTitleBarHidden", data_type="bool", def_value=False
        )

    @title_bar_hidden.setter
    def title_bar_hidden(self, value: Optional[bool]):
        self.page._set_attr("windowTitleBarHidden", value)

    # title_bar_buttons_hidden
    @property
    def title_bar_buttons_hidden(self) -> bool:
        return self.page._get_attr(
            "windowTitleBarButtonsHidden", data_type="bool", def_value=False
        )

    @title_bar_buttons_hidden.setter
    def title_bar_buttons_hidden(self, value: Optional[bool]):
        self.page._set_attr("windowTitleBarButtonsHidden", value)

    # skip_task_bar
    @property
    def skip_task_bar(self) -> bool:
        return self.page._get_attr(
            "windowSkipTaskBar", data_type="bool", def_value=False
        )

    @skip_task_bar.setter
    def skip_task_bar(self, value: Optional[bool]):
        self.page._set_attr("windowSkipTaskBar", value)

    # frameless
    @property
    def frameless(self) -> bool:
        return self.page._get_attr("windowFrameless", data_type="bool", def_value=False)

    @frameless.setter
    def frameless(self, value: Optional[bool]):
        self.page._set_attr("windowFrameless", value)

    # progress_bar
    @property
    def progress_bar(self) -> OptionalNumber:
        return self.page._get_attr("windowProgressBar")

    @progress_bar.setter
    def progress_bar(self, value: OptionalNumber):
        self.page._set_attr("windowProgressBar", value)

    # focused
    @property
    def focused(self) -> bool:
        return self.page._get_attr("windowFocused", data_type="bool", def_value=True)

    @focused.setter
    def focused(self, value: Optional[bool]):
        self.page._set_attr("windowFocused", value)

    # visible
    @property
    def visible(self) -> bool:
        return self.page._get_attr("windowVisible", data_type="bool", def_value=True)

    @visible.setter
    def visible(self, value: Optional[bool]):
        self.page._set_attr("windowVisible", value)

    # always_on_bottom
    @property
    def always_on_bottom(self) -> bool:
        return self.page._get_attr(
            "windowAlwaysOnBottom", data_type="bool", def_value=False
        )

    @always_on_bottom.setter
    def always_on_bottom(self, value: Optional[bool]):
        self.page._set_attr("windowAlwaysOnBottom", value)

    # wait_until_ready_to_show
    @property
    def wait_until_ready_to_show(self) -> bool:
        return self.page._get_attr(
            "windowWaitUntilReadyToShow", data_type="bool", def_value=False
        )

    @wait_until_ready_to_show.setter
    def wait_until_ready_to_show(self, value: Optional[bool]):
        self.page._set_attr("windowWaitUntilReadyToShow", value)

    # shadow
    @property
    def shadow(self) -> bool:
        return self.page._get_attr("windowShadow", data_type="bool", def_value=False)

    @shadow.setter
    def shadow(self, value: Optional[bool]):
        self.page._set_attr("windowShadow", value)

    # alignment
    @property
    def alignment(self) -> Optional[Alignment]:
        return self.__alignment

    @alignment.setter
    def alignment(self, value: Optional[Alignment]):
        self.__alignment = value

    # badge_label
    @property
    def badge_label(self) -> Optional[str]:
        return self.page._get_attr("windowBadgeLabel")

    @badge_label.setter
    def badge_label(self, value: Optional[str]):
        self.page._set_attr("windowBadgeLabel", value)

    # icon
    @property
    def icon(self) -> Optional[str]:
        return self.page._get_attr("windowIcon")

    @icon.setter
    def icon(self, value: Optional[str]):
        self.page._set_attr("windowIcon", value)

    # ignore_mouse_events
    @property
    def ignore_mouse_events(self) -> bool:
        return self.page._get_attr(
            "windowIgnoreMouseEvents", data_type="bool", def_value=False
        )

    @ignore_mouse_events.setter
    def ignore_mouse_events(self, value: Optional[bool]):
        self.page._set_attr("windowIgnoreMouseEvents", value)

    # Methods
    def destroy(self):
        self.page._set_attr("windowDestroy", True)
        self.page.update()

    def center(self) -> None:
        self.page._set_attr("windowCenter", str(time.time()))
        self.page.update()

    def close(self) -> None:
        self.page._set_attr("windowClose", str(time.time()))
        self.page.update()

    def to_front(self) -> None:
        self.page._invoke_method("windowToFront")

    def start_dragging(self) -> None:
        self.page._invoke_method("windowStartDragging")

    # Events
    # on_event
    @property
    def on_event(self) -> OptionalEventCallable["WindowEvent"]:
        return self.__on_event.handler

    @on_event.setter
    def on_event(
        self,
        handler: OptionalEventCallable["WindowEvent"],
    ):
        self.__on_event.handler = handler


class Page(AdaptiveControl):
    """
    Page is a container for `View` (https://flet.dev/docs/controls/view) controls.

    A page instance and the root view are automatically created when a new user session started.

    Example:

    ```
    import flet as ft


    def main(page: ft.Page):
        page.title = "New page"
        page.add(ft.Text("Hello"))


    ft.app(target=main)
    ```

    -----

    Online docs: https://flet.dev/docs/controls/page
    """

    def __init__(
        self,
        conn: Connection,
        session_id,
        loop: asyncio.AbstractEventLoop,
        executor: Optional[ThreadPoolExecutor] = None,
    ) -> None:
        Control.__init__(self)
        self._id = "page"
        self._Control__uid = "page"
        self.__window = Window(self)
        self.__browser_context_menu = BrowserContextMenu(self)
        self.__conn = conn
        self.__next_control_id = 1
        self.__snapshot: Dict[str, Dict[str, Any]] = {}
        self.__expires_at = None
        self.__query: QueryString = QueryString(page=self)  # Querystring
        self._session_id = session_id
        self.__loop = loop
        self.__executor = executor
        self._index = {self._Control__uid: self}  # index with all page controls

        self.__lock = threading.Lock() if not is_pyodide() else NopeLock()

        self.__views = [View()]
        self.__default_view = self.__views[0]
        self._controls = self.__default_view.controls

        self.__fonts: Optional[Dict[str, str]] = None
        self.__offstage = Offstage()
        self.__theme = None
        self.__dark_theme = None
        self.__locale_configuration = None
        self.__theme_mode = ThemeMode.SYSTEM  # Default Theme Mode
        self.__pubsub: PubSubClient = PubSubClient(conn.pubsubhub, session_id)
        self.__client_storage: ClientStorage = ClientStorage(self)
        self.__session_storage: SessionStorage = SessionStorage(self)
        self.__authorization: Optional[Authorization] = None

        self.__on_close = EventHandler()
        self._add_event_handler("close", self.__on_close.get_handler())
        self.__on_platform_brightness_change = EventHandler()
        self._add_event_handler(
            "platformBrightnessChange",
            self.__on_platform_brightness_change.get_handler(),
        )

        self.__on_app_lifecycle_state_change = EventHandler(
            lambda e: AppLifecycleStateChangeEvent(e)
        )
        self._add_event_handler(
            "app_lifecycle_state_change",
            self.__on_app_lifecycle_state_change.get_handler(),
        )
        self.__on_resized = EventHandler(lambda e: WindowResizeEvent(e))
        self._add_event_handler("resized", self.__on_resized.get_handler())

        self.__last_route = None

        # authorize/login/logout

        self.__on_login = EventHandler()
        self._add_event_handler("authorize", self.__on_authorize_async)
        self.__on_logout = EventHandler()

        # route_change

        def convert_route_change_event(e):
            if self.__last_route == e.data:
                return None  # avoid duplicate calls
            self.__last_route = e.data
            self._set_attr("route", e.data, False)
            self.query()  # Update query url (required when manually changed from browser)
            return RouteChangeEvent(route=e.data)

        self.__on_route_change: EventHandler = EventHandler(convert_route_change_event)
        self._add_event_handler("route_change", self.__on_route_change.get_handler())

        def convert_view_pop_event(e):
            # e.data contains route name
            view = next((v for v in self.views if v.route == e.data), None)
            return ViewPopEvent(view=view) if view in self.views else None

        self.__on_view_pop = EventHandler(convert_view_pop_event)
        self._add_event_handler("view_pop", self.__on_view_pop.get_handler())

        def convert_keyboard_event(e):
            d = json.loads(e.data)
            return KeyboardEvent(**d)

        self.__on_keyboard_event = EventHandler(convert_keyboard_event)
        self._add_event_handler(
            "keyboard_event", self.__on_keyboard_event.get_handler()
        )

        def convert_page_media_change_event(e):
            d = json.loads(e.data)
            return PageMediaData(**d)

        self.__on_page_media_change_event = EventHandler(
            convert_page_media_change_event
        )
        self._add_event_handler(
            "mediaChange", self.__on_page_media_change_event.get_handler()
        )

        self.__method_calls: Dict[str, Union[threading.Event, asyncio.Event]] = {}
        self.__method_call_results: Dict[
            Union[threading.Event, asyncio.Event], tuple[Optional[str], Optional[str]]
        ] = {}
        self._add_event_handler("invoke_method_result", self.__on_invoke_method_result)

        self.__on_connect = EventHandler()
        self._add_event_handler("connect", self.__on_connect.get_handler())
        self.__on_disconnect = EventHandler()
        self._add_event_handler("disconnect", self.__on_disconnect.get_handler())
        self.__on_error = EventHandler()
        self._add_event_handler("error", self.__on_error.get_handler())

        _session_page.set(self)

    def get_control(self, id: str) -> Control:
        return self._index.get(id)

    def before_update(self) -> None:
        super().before_update()
        self._set_attr_json("fonts", self.__fonts)
        self._set_attr_json("theme", self.__theme)
        self._set_attr_json("localeConfiguration", self.__locale_configuration)
        self._set_attr_json("darkTheme", self.__dark_theme)
        self._set_attr_json("windowAlignment", self.__window.alignment)

    def _get_control_name(self):
        return "page"

    def _get_children(self):
        children = []
        children.extend(self.__views)
        children.append(self.__offstage)
        return children

    def get_next_control_id(self) -> int:
        r = self.__next_control_id
        self.__next_control_id += 1
        return r

    async def fetch_page_details_async(self) -> None:
        assert self.__conn
        props = [
            "route",
            "pwa",
            "web",
            "debug",
            "platform",
            "platformBrightness",
            "media",
            "width",
            "height",
            "windowWidth",
            "windowHeight",
            "windowTop",
            "windowLeft",
            "clientIP",
            "clientUserAgent",
        ]
        values = (
            self.__conn.send_commands(
                self._session_id,
                [Command(0, "get", ["page", prop]) for prop in props],
            )
        ).results
        for i in range(len(props)):
            self._set_attr(props[i], values[i], False)

    async def _connect(self, conn: Connection) -> None:
        _session_page.set(self)
        self.__conn = conn
        self.__expires_at = None
        await self.on_event_async(Event("page", "connect", ""))

    async def _disconnect(self, session_timeout_seconds: int) -> None:
        self.__expires_at = datetime.now(timezone.utc) + timedelta(
            seconds=session_timeout_seconds
        )
        await self.on_event_async(Event("page", "disconnect", ""))

    def update(self, *controls) -> None:
        with self.__lock:
            if len(controls) == 0:
                r = self.__update(self)
            else:
                r = self.__update(*controls)
        self.__handle_mount_unmount(*r)

    def add(self, *controls: Control) -> None:
        with self.__lock:
            self._controls.extend(controls)
            r = self.__update(self)
        self.__handle_mount_unmount(*r)

    def insert(self, at: int, *controls: Control) -> None:
        with self.__lock:
            n = at
            for control in controls:
                self._controls.insert(n, control)
                n += 1
            r = self.__update(self)
        self.__handle_mount_unmount(*r)

    def remove(self, *controls: Control) -> None:
        with self.__lock:
            for control in controls:
                self._controls.remove(control)
            r = self.__update(self)
        self.__handle_mount_unmount(*r)

    def remove_at(self, index: int) -> None:
        with self.__lock:
            self._controls.pop(index)
            r = self.__update(self)
        self.__handle_mount_unmount(*r)

    def clean(self) -> None:
        self._clean(self)
        self._controls.clear()

    def _clean(self, control: Control) -> None:
        with self.__lock:
            control._previous_children.clear()
            assert control.uid is not None
            removed_controls = []
            for child in control._get_children():
                removed_controls.extend(
                    self._remove_control_recursively(self.index, child)
                )
            self._send_command("clean", [control.uid])
            for c in removed_controls:
                c.will_unmount()

    def _close(self) -> None:
        self.__pubsub.unsubscribe_all()
        removed_controls = self._remove_control_recursively(self.index, self)
        for c in removed_controls:
            c.will_unmount()
            c._dispose()
        self._controls.clear()
        self._previous_children.clear()
        self.__on_view_pop = None
        self.__client_storage = None
        self.__session_storage = None
        self.__conn = None

    def __update(self, *controls: Control) -> Tuple[List[Control], List[Control]]:
        if not self.__conn:
            raise PageDisconnectedException("Page has been disconnected")
        commands, added_controls, removed_controls = self.__prepare_update(*controls)
        self.__validate_controls_page(added_controls)
        results = self.__conn.send_commands(self._session_id, commands).results
        self.__update_control_ids(added_controls, results)
        return added_controls, removed_controls

    def __prepare_update(
        self, *controls: Control
    ) -> Tuple[List[Any], List[Control], List[Control]]:
        added_controls = []
        removed_controls = []
        commands = []

        # build commands

        for control in controls:
            control.build_update_commands(
                self._index, commands, added_controls, removed_controls
            )
        return commands, added_controls, removed_controls

    def __validate_controls_page(self, added_controls: List[Control]) -> None:
        for ctrl in added_controls:
            if ctrl.page and ctrl.page != self:
                raise Exception(
                    f"Control has already been added to another page: {ctrl}"
                )

    def __update_control_ids(
        self, added_controls: List[Control], results: List[Any]
    ) -> None:
        if len(results) > 0:
            n = 0
            for line in results:
                for id in line.split(" "):
                    added_controls[n]._Control__uid = id

                    # add to index
                    self._index[id] = added_controls[n]

                    n += 1

    def __handle_mount_unmount(self, added_controls, removed_controls) -> None:
        for ctrl in removed_controls:
            ctrl.will_unmount()
            ctrl.parent = None  # remove parent reference
            ctrl.page = None
        for ctrl in added_controls:
            ctrl.did_mount()

    def error(self, message: str = "") -> None:
        with self.__lock:
            self._send_command("error", [message])

    async def on_event_async(self, e: Event) -> None:
        logger.debug(f"page.on_event_async: {e.target} {e.name} {e.data}")

        if e.target == "page" and e.name == "change":
            with self.__lock:
                self.__on_page_change_event(e.data)
        elif e.target in self._index:
            ce = ControlEvent(e.target, e.name, e.data, self._index[e.target], self)
            handler = self._index[e.target].event_handlers.get(e.name)
            if handler:
                if asyncio.iscoroutinefunction(handler):
                    await handler(ce)
                else:
                    self.run_thread(handler, ce)

    def __on_page_change_event(self, data: str) -> None:
        for props in json.loads(data):
            id = props["i"]
            if id in self._index:
                for name in props:
                    if name != "i":
                        self._index[id]._set_attr(name, props[name], dirty=False)
                        if id in self.__snapshot:
                            self.__snapshot[id][name] = props[name]

    def run_task(
        self,
        handler: Callable[InputT, Awaitable[RetT]],
        *args: InputT.args,
        **kwargs: InputT.kwargs,
    ) -> Future[RetT]:
        _session_page.set(self)
        assert asyncio.iscoroutinefunction(handler)

        future = asyncio.run_coroutine_threadsafe(handler(*args, **kwargs), self.__loop)

        def _on_completion(f):
            try:
                exception = f.exception()
                if exception:
                    raise exception
            except CancelledError:
                pass

        future.add_done_callback(_on_completion)

        return future

    def __context_wrapper(self, handler: Callable[..., Any]) -> Wrapper:
        def wrapper(*args):
            _session_page.set(self)
            handler(*args)

        return wrapper

    def run_thread(
        self,
        handler: Callable[InputT, Any],
        *args: InputT.args,
        **kwargs: InputT.kwargs,
    ) -> None:
        handler_with_context = self.__context_wrapper(handler)
        if is_pyodide():
            handler_with_context(*args, **kwargs)
        else:
            assert self.__loop
            self.__loop.call_soon_threadsafe(
                self.__loop.run_in_executor,
                self.__executor,
                partial(handler_with_context, *args, **kwargs),
            )

    def go(
        self, route: str, skip_route_change_event: bool = False, **kwargs: Any
    ) -> None:
        self.route = route if not kwargs else route + self.query.post(kwargs)

        if not skip_route_change_event:
            self.run_task(
                self.__on_route_change.get_handler(),
                ControlEvent(
                    target="page",
                    name="route_change",
                    data=self.route,
                    page=self,
                    control=self,
                ),
            )
        self.update()
        self.query()  # Update query url (required when using go)

    def get_upload_url(self, file_name: str, expires: int) -> str:
        r = self._send_command(
            "getUploadUrl", attrs={"file": file_name, "expires": str(expires)}
        )
        if r.error:
            raise Exception(r.error)
        return r.result

    def login(
        self,
        provider: OAuthProvider,
        fetch_user: Optional[bool] = True,
        fetch_groups: Optional[bool] = False,
        scope: Optional[List[str]] = None,
        saved_token: Optional[str] = None,
        on_open_authorization_url: Optional[Callable[[str], None]] = None,
        complete_page_html: Optional[str] = None,
        redirect_to_page: Optional[bool] = False,
        authorization: Type[AT] = Authorization,
    ) -> AT:
        self.__authorization = authorization(
            provider,
            fetch_user=fetch_user,
            fetch_groups=fetch_groups,
            scope=scope,
        )
        if saved_token is None:
            authorization_url, state = self.__authorization.get_authorization_data()
            auth_attrs = {"state": state}
            if complete_page_html:
                auth_attrs["completePageHtml"] = complete_page_html
            if redirect_to_page:
                up = urlparse(provider.redirect_url)
                auth_attrs["completePageUrl"] = up._replace(
                    path=self.__conn.page_name
                ).geturl()
            result = self._send_command("oauthAuthorize", attrs=auth_attrs)
            if result.error != "":
                raise Exception(result.error)
            if on_open_authorization_url:
                on_open_authorization_url(authorization_url)
            else:
                self.launch_url(
                    authorization_url, "flet_oauth_signin", web_popup_window=self.web
                )
        else:
            self.__authorization.dehydrate_token(saved_token)
            self.run_task(
                self.__on_login.get_handler(),
                LoginEvent(
                    error="",
                    error_description="",
                    page=self,
                    control=self,
                    target="page",
                    name="on_login",
                    data="",
                ),
            )
        return self.__authorization

    async def login_async(
        self,
        provider: OAuthProvider,
        fetch_user: Optional[bool] = True,
        fetch_groups: Optional[bool] = False,
        scope: Optional[List[str]] = None,
        saved_token: Optional[str] = None,
        on_open_authorization_url: Optional[
            Callable[[str], Coroutine[Any, Any, None]]
        ] = None,
        complete_page_html: Optional[str] = None,
        redirect_to_page: Optional[bool] = False,
        authorization: Type[AT] = Authorization,
    ) -> AT:
        self.__authorization = authorization(
            provider,
            fetch_user=fetch_user,
            fetch_groups=fetch_groups,
            scope=scope,
        )
        if saved_token is None:
            authorization_url, state = self.__authorization.get_authorization_data()
            auth_attrs = {"state": state}
            if complete_page_html:
                auth_attrs["completePageHtml"] = complete_page_html
            if redirect_to_page:
                up = urlparse(provider.redirect_url)
                auth_attrs["completePageUrl"] = up._replace(
                    path=f"{self.__conn.page_name}{self.route}"
                ).geturl()
            result = self._send_command("oauthAuthorize", attrs=auth_attrs)
            if result.error != "":
                raise Exception(result.error)
            if on_open_authorization_url:
                await on_open_authorization_url(authorization_url)
            else:
                self.launch_url(
                    authorization_url, "flet_oauth_signin", web_popup_window=self.web
                )
        else:
            await self.__authorization.dehydrate_token_async(saved_token)
            self.run_task(
                self.__on_login.get_handler(),
                LoginEvent(
                    error="",
                    error_description="",
                    page=self,
                    control=self,
                    target="page",
                    name="on_login",
                    data="",
                ),
            )
        return self.__authorization

    async def _authorize_callback_async(self, data: str) -> None:
        await self.on_event_async(Event("page", "authorize", json.dumps(data)))

    async def __on_authorize_async(self, e) -> None:
        assert self.__authorization
        d = json.loads(e.data)
        state = d.get("state")
        assert state == self.__authorization.state

        if not self.web:
            if self.platform in ["ios", "android"]:
                # close web view on mobile
                self.close_in_app_web_view()
            else:
                # activate desktop window
                self.window.to_front()
        login_evt = LoginEvent(
            error=d.get("error"),
            error_description=d.get("error_description"),
            page=self,
            control=self,
            target="page",
            name="on_login",
            data="",
        )
        if not login_evt.error:
            # perform token request

            code = d.get("code")
            assert code not in [None, ""]
            try:
                await self.__authorization.request_token_async(code)
            except Exception as ex:
                login_evt.error = str(ex)
        self.run_task(
            self.__on_login.get_handler(),
            login_evt,
        )

    def logout(self) -> None:
        self.__authorization = None
        self.run_task(
            self.__on_logout.get_handler(),
            ControlEvent(
                target="page", name="logout", data="", control=self, page=self
            ),
        )

    def _send_command(
        self,
        name: str,
        values: Optional[List[str]] = None,
        attrs: Optional[Dict[str, str]] = None,
    ):
        return self.__conn.send_command(
            self._session_id,
            Command(
                indent=0,
                name=name,
                values=values if values is not None else [],
                attrs=attrs or {},
            ),
        )

    def set_clipboard(self, value: str, wait_timeout: Optional[float] = 10) -> None:
        self._invoke_method("setClipboard", {"data": value}, wait_timeout=wait_timeout)

    def get_clipboard(self, wait_timeout: Optional[float] = 10) -> Optional[str]:
        return self._invoke_method(
            "getClipboard", wait_for_result=True, wait_timeout=wait_timeout
        )

    async def get_clipboard_async(
        self, wait_timeout: Optional[float] = 10
    ) -> Optional[str]:
        return await self._invoke_method_async(
            "getClipboard", wait_for_result=True, wait_timeout=wait_timeout
        )

    def launch_url(
        self,
        url: str,
        web_window_name: Optional[str] = None,
        web_popup_window: Optional[bool] = False,
        window_width: Optional[int] = None,
        window_height: Optional[int] = None,
    ) -> None:
        args = {"url": url}
        if web_window_name:
            args["web_window_name"] = web_window_name
        if web_popup_window:
            args["web_popup_window"] = str(web_popup_window)
        if window_width:
            args["window_width"] = str(window_width)
        if window_height:
            args["window_height"] = str(window_height)
        self._invoke_method("launchUrl", args)

    def can_launch_url(self, url: str) -> bool:
        args = {"url": url}
        return self._invoke_method("canLaunchUrl", args, wait_for_result=True) == "true"

    async def can_launch_url_async(self, url: str) -> bool:
        args = {"url": url}
        return (
            await self._invoke_method_async("canLaunchUrl", args, wait_for_result=True)
            == "true"
        )

    def close_in_app_web_view(self) -> None:
        self._invoke_method("closeInAppWebView")

    def scroll_to(
        self,
        offset: Optional[float] = None,
        delta: Optional[float] = None,
        key: Optional[str] = None,
        duration: Optional[int] = None,
        curve: Optional[AnimationCurve] = None,
    ) -> None:
        self.__default_view.scroll_to(
            offset=offset, delta=delta, key=key, duration=duration, curve=curve
        )

    def _invoke_method(
        self,
        method_name: str,
        arguments: Optional[Dict[str, str]] = None,
        control_id: Optional[str] = "",
        wait_for_result: Optional[bool] = False,
        wait_timeout: Optional[float] = 5,
    ) -> Optional[str]:
        method_id = uuid.uuid4().hex

        # register callback
        evt: Optional[threading.Event] = None
        if wait_for_result:
            evt = threading.Event()
            self.__method_calls[method_id] = evt

        # call method
        result = self._send_command(
            "invokeMethod", values=[method_id, method_name, control_id], attrs=arguments
        )

        if result.error != "":
            if wait_for_result:
                del self.__method_calls[method_id]
            raise Exception(result.error)
        if not wait_for_result:
            return
        assert evt is not None

        if not evt.wait(wait_timeout):
            del self.__method_calls[method_id]
            raise TimeoutError(
                f"Timeout waiting for invokeMethod {method_name}({arguments}) call"
            )

        result, err = self.__method_call_results.pop(evt)
        if err:
            raise Exception(err)
        if result is None or result == "null":
            return None
        return result

    async def _invoke_method_async(
        self,
        method_name: str,
        arguments: Optional[Dict[str, str]] = None,
        control_id: Optional[str] = "",
        wait_for_result: Optional[bool] = False,
        wait_timeout: Optional[float] = 5,
    ) -> Optional[str]:
        method_id = uuid.uuid4().hex

        # register callback
        evt: Optional[asyncio.Event] = None
        if wait_for_result:
            evt = asyncio.Event()
            self.__method_calls[method_id] = evt

        # call method
        result = self._send_command(
            "invokeMethod", values=[method_id, method_name, control_id], attrs=arguments
        )

        if result.error != "":
            if wait_for_result:
                del self.__method_calls[method_id]
            raise Exception(result.error)
        if not wait_for_result:
            return
        assert evt is not None

        try:
            await asyncio.wait_for(evt.wait(), timeout=wait_timeout)
        except TimeoutError:
            del self.__method_calls[method_id]
            raise TimeoutError(
                f"Timeout waiting for invokeMethod {method_name}({arguments}) call"
            )

        result, err = self.__method_call_results.pop(evt)
        if err:
            raise Exception(err)
        if result == "null":
            return None
        return result

    def __on_invoke_method_result(self, e) -> None:
        d = json.loads(e.data)
        result = InvokeMethodResults(**d)
        evt = self.__method_calls.pop(result.method_id, None)
        if evt is None:
            return
        self.__method_call_results[evt] = (result.result, result.error)
        evt.set()

    def open(self, control: Control) -> None:
        if not hasattr(control, "open"):
            raise ValueError(f"{control.__class__.__qualname__} has no open attribute")

        control.open = True

        if isinstance(control, NavigationDrawer):
            if control.position == NavigationDrawerPosition.END:
                if self.end_drawer != control:
                    self.end_drawer = control
                    self.update()
            else:
                if self.drawer != control:
                    self.drawer = control
                    self.update()
        else:
            if control not in self.__offstage.controls:
                self.__offstage.controls.append(control)
                self.__offstage.update()

        control.update()

    @staticmethod
    def close(control: Control) -> None:
        if hasattr(control, "open"):
            control.open = False
            control.update()
        else:
            raise ValueError(f"{control.__class__.__qualname__} has no open attribute")

    # query
    @property
    def query(self) -> QueryString:
        return self.__query

    # url
    @property
    def url(self) -> Optional[str]:
        return self.__conn.page_url

    # name
    @property
    def name(self) -> str:
        return self.__conn.page_name

    # connection
    @property
    def connection(self) -> Optional[Connection]:
        return self.__conn

    # snapshot
    @property
    def snapshot(self) -> Dict[str, Dict[str, Any]]:
        return self.__snapshot

    # loop
    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self.__loop

    # executor
    @property
    def executor(self) -> Optional[ThreadPoolExecutor]:
        return self.__executor

    # expires_at
    @property
    def expires_at(self) -> Optional[datetime]:
        return self.__expires_at

    # index
    @property
    def index(self) -> "Dict[str, Page]":
        return self._index

    # session_id
    @property
    def session_id(self) -> Any:
        return self._session_id

    # auth
    @property
    def auth(self) -> Optional[Authorization]:
        return self.__authorization

    # pubsub
    @property
    def pubsub(self) -> PubSubClient:
        return self.__pubsub

    # overlay
    @property
    def overlay(self) -> List[Control]:
        return self.__offstage.controls

    # title
    @property
    def title(self) -> str:
        return self._get_attr("title")

    @title.setter
    def title(self, value: str):
        self._set_attr("title", value)

    # route
    @property
    def route(self) -> str:
        return self._get_attr("route")

    @route.setter
    def route(self, value: str):
        self._set_attr("route", value)

    # pwa
    @property
    def pwa(self) -> bool:
        return self._get_attr("pwa", data_type="bool", def_value=False)

    # web
    @property
    def web(self) -> bool:
        return cast(bool, self._get_attr("web", data_type="bool", def_value=False))

    # debug
    @property
    def debug(self) -> bool:
        return cast(bool, self._get_attr("debug", data_type="bool", def_value=False))

    # platform
    @property
    def platform(self) -> PagePlatform:
        return PagePlatform(self._get_attr("platform"))

    @platform.setter
    def platform(self, value: PagePlatform):
        self._set_attr(
            "platform", value.value if isinstance(value, PagePlatform) else value
        )

    # platform_brightness
    @property
    def platform_brightness(self) -> Brightness:
        brightness = self._get_attr("platformBrightness")
        assert brightness
        return Brightness(brightness)

    # media
    @property
    def media(self) -> Optional["PageMediaData"]:
        m = self._get_attr("media")
        if not isinstance(m, str):
            return None
        d = json.loads(m)
        return PageMediaData(**d)

    # client_ip
    @property
    def client_ip(self):
        return self._get_attr("clientIP")

    # client_user_agent
    @property
    def client_user_agent(self):
        return self._get_attr("clientUserAgent")

    # fonts
    @property
    def fonts(self) -> Optional[Dict[str, str]]:
        return self.__fonts

    @fonts.setter
    def fonts(self, value: Optional[Dict[str, str]]):
        self.__fonts = value

    # views
    @property
    def views(self) -> List[View]:
        return self.__views

    # controls
    @property
    def controls(self) -> Optional[List[Control]]:
        return self.__default_view.controls

    @controls.setter
    def controls(self, value: Optional[Sequence[Control]]):
        self.__default_view.controls = list(value) if value is not None else []

    # appbar
    @property
    def appbar(self) -> Union[AppBar, CupertinoAppBar, None]:
        return self.__default_view.appbar

    @appbar.setter
    def appbar(self, value: Union[AppBar, CupertinoAppBar, None]):
        self.__default_view.appbar = value

    # bottom_appbar
    @property
    def bottom_appbar(self) -> Optional[BottomAppBar]:
        return self.__default_view.bottom_appbar

    @bottom_appbar.setter
    def bottom_appbar(self, value: Optional[BottomAppBar]):
        self.__default_view.bottom_appbar = value

    # navigation_bar
    @property
    def navigation_bar(self) -> Optional[Union[NavigationBar, CupertinoNavigationBar]]:
        return self.__default_view.navigation_bar

    @navigation_bar.setter
    def navigation_bar(
        self,
        value: Optional[Union[NavigationBar, CupertinoNavigationBar]],
    ):
        self.__default_view.navigation_bar = value

    # drawer
    @property
    def drawer(self) -> Optional[NavigationDrawer]:
        return self.__default_view.drawer

    @drawer.setter
    def drawer(self, value: Optional[NavigationDrawer]):
        self.__default_view.drawer = value

    # end_drawer
    @property
    def end_drawer(self) -> Optional[NavigationDrawer]:
        return self.__default_view.end_drawer

    @end_drawer.setter
    def end_drawer(self, value: Optional[NavigationDrawer]):
        self.__default_view.end_drawer = value

    # decoration
    @property
    def decoration(self) -> Optional[BoxDecoration]:
        return self.__default_view.decoration

    @decoration.setter
    def decoration(self, value: Optional[BoxDecoration]):
        self.__default_view.decoration = value

    # foreground_decoration
    @property
    def foreground_decoration(self) -> Optional[BoxDecoration]:
        return self.__default_view.foreground_decoration

    @foreground_decoration.setter
    def foreground_decoration(self, value: Optional[BoxDecoration]):
        self.__default_view.foreground_decoration = value

    # floating_action_button
    @property
    def floating_action_button(self) -> Optional[FloatingActionButton]:
        return self.__default_view.floating_action_button

    @floating_action_button.setter
    def floating_action_button(self, value: Optional[FloatingActionButton]):
        self.__default_view.floating_action_button = value

    # floating_action_button_location
    @property
    def floating_action_button_location(
        self,
    ) -> Union[FloatingActionButtonLocation, OffsetValue]:
        return self.__default_view.floating_action_button_location

    @floating_action_button_location.setter
    def floating_action_button_location(
        self, value: Union[FloatingActionButtonLocation, OffsetValue]
    ):
        self.__default_view.floating_action_button_location = value

    # horizontal_alignment
    @property
    def horizontal_alignment(self) -> CrossAxisAlignment:
        return self.__default_view.horizontal_alignment

    @horizontal_alignment.setter
    def horizontal_alignment(self, value: CrossAxisAlignment):
        self.__default_view.horizontal_alignment = value

    # vertical_alignment
    @property
    def vertical_alignment(self) -> MainAxisAlignment:
        return self.__default_view.vertical_alignment

    @vertical_alignment.setter
    def vertical_alignment(self, value: MainAxisAlignment):
        self.__default_view.vertical_alignment = value

    # window
    @property
    def window(self) -> Window:
        return self.__window

    # browser_context_menu
    @property
    def browser_context_menu(self) -> BrowserContextMenu:
        return self.__browser_context_menu

    # spacing
    @property
    def spacing(self) -> OptionalNumber:
        return self.__default_view.spacing

    @spacing.setter
    def spacing(self, value: OptionalNumber):
        self.__default_view.spacing = value

    # padding
    @property
    def padding(self) -> Optional[PaddingValue]:
        return self.__default_view.padding

    @padding.setter
    def padding(self, value: Optional[PaddingValue]):
        self.__default_view.padding = value

    # bgcolor
    @property
    def bgcolor(self) -> Optional[ColorValue]:
        return self.__default_view.bgcolor

    @bgcolor.setter
    def bgcolor(self, value: Optional[ColorValue]):
        self.__default_view.bgcolor = value

    # scroll
    @property
    def scroll(self) -> Optional[ScrollMode]:
        return self.__default_view.scroll

    @scroll.setter
    def scroll(self, value: Optional[ScrollMode]):
        self.__default_view.scroll = value

    # auto_scroll
    @property
    def auto_scroll(self) -> bool:
        return self.__default_view.auto_scroll

    @auto_scroll.setter
    def auto_scroll(self, value: Optional[bool]):
        self.__default_view.auto_scroll = value

    # client_storage
    @property
    def client_storage(self) -> ClientStorage:
        return self.__client_storage

    # session_storage
    @property
    def session(self) -> SessionStorage:
        return self.__session_storage

    # theme_mode
    @property
    def theme_mode(self) -> Optional[ThemeMode]:
        return self.__theme_mode

    @theme_mode.setter
    def theme_mode(self, value: Optional[ThemeMode]):
        self.__theme_mode = value
        self._set_attr(
            "themeMode", value.value if isinstance(value, ThemeMode) else value
        )

    # theme
    @property
    def theme(self) -> Optional[Theme]:
        return self.__theme

    @theme.setter
    def theme(self, value: Optional[Theme]):
        self.__theme = value

    # dark_theme
    @property
    def dark_theme(self) -> Optional[Theme]:
        return self.__dark_theme

    @dark_theme.setter
    def dark_theme(self, value: Optional[Theme]):
        self.__dark_theme = value

    # locale_configuration
    @property
    def locale_configuration(self) -> Optional[LocaleConfiguration]:
        return self.__locale_configuration

    @locale_configuration.setter
    def locale_configuration(self, value: Optional[LocaleConfiguration]):
        self.__locale_configuration = value

    # rtl
    @property
    def rtl(self) -> bool:
        return self._get_attr("rtl", data_type="bool", def_value=False)

    @rtl.setter
    def rtl(self, value: Optional[bool]):
        self._set_attr("rtl", value)

    # show_semantics_debugger
    @property
    def show_semantics_debugger(self) -> bool:
        return self._get_attr(
            "showSemanticsDebugger", data_type="bool", def_value=False
        )

    @show_semantics_debugger.setter
    def show_semantics_debugger(self, value: Optional[bool]):
        self._set_attr("showSemanticsDebugger", value)

    # width
    @property
    def width(self) -> OptionalNumber:
        w = self._get_attr("width")
        return float(w) if w else 0

    # height
    @property
    def height(self) -> OptionalNumber:
        h = self._get_attr("height")
        return float(h) if h else 0

    # on_scroll_interval
    @property
    def on_scroll_interval(self) -> OptionalNumber:
        return self.__default_view.on_scroll_interval

    @on_scroll_interval.setter
    def on_scroll_interval(self, value: OptionalNumber):
        self.__default_view.on_scroll_interval = value

    # on_close
    @property
    def on_close(self) -> OptionalControlEventCallable:
        return self.__on_close.handler

    @on_close.setter
    def on_close(self, handler: OptionalControlEventCallable):
        self.__on_close.handler = handler

    @property
    def on_resized(self) -> OptionalEventCallable["WindowResizeEvent"]:
        return self.__on_resized.handler

    @on_resized.setter
    def on_resized(self, handler: OptionalEventCallable["WindowResizeEvent"]):
        self.__on_resized.handler = handler

    # on_platform_brightness_change
    @property
    def on_platform_brightness_change(self) -> OptionalControlEventCallable:
        return self.__on_platform_brightness_change.handler

    @on_platform_brightness_change.setter
    def on_platform_brightness_change(self, handler: OptionalControlEventCallable):
        self.__on_platform_brightness_change.handler = handler

    # on_app_lifecycle_change
    @property
    def on_app_lifecycle_state_change(
        self,
    ) -> OptionalEventCallable["AppLifecycleStateChangeEvent"]:
        return self.__on_app_lifecycle_state_change.handler

    @on_app_lifecycle_state_change.setter
    def on_app_lifecycle_state_change(
        self, handler: OptionalEventCallable["AppLifecycleStateChangeEvent"]
    ):
        self.__on_app_lifecycle_state_change.handler = handler

    # on_route_change
    @property
    def on_route_change(self) -> OptionalEventCallable["RouteChangeEvent"]:
        return self.__on_route_change.handler

    @on_route_change.setter
    def on_route_change(self, handler: OptionalEventCallable["RouteChangeEvent"]):
        self.__on_route_change.handler = handler

    # on_view_pop
    @property
    def on_view_pop(self) -> OptionalEventCallable["ViewPopEvent"]:
        return self.__on_view_pop.handler

    @on_view_pop.setter
    def on_view_pop(self, handler: OptionalEventCallable["ViewPopEvent"]):
        self.__on_view_pop.handler = handler

    # on_keyboard_event
    @property
    def on_keyboard_event(self) -> OptionalEventCallable["KeyboardEvent"]:
        return self.__on_keyboard_event.handler

    @on_keyboard_event.setter
    def on_keyboard_event(self, handler: OptionalEventCallable["KeyboardEvent"]):
        self.__on_keyboard_event.handler = handler
        self._set_attr("onKeyboardEvent", True if handler else None)

    # on_media_change
    @property
    def on_media_change(self) -> OptionalEventCallable["PageMediaData"]:
        return self.__on_page_media_change_event.handler

    @on_media_change.setter
    def on_media_change(self, handler: OptionalEventCallable["PageMediaData"]):
        self.__on_page_media_change_event.handler = handler

    # on_connect
    @property
    def on_connect(self) -> OptionalControlEventCallable:
        return self.__on_connect.handler

    @on_connect.setter
    def on_connect(self, handler: OptionalControlEventCallable):
        self.__on_connect.handler = handler

    # on_disconnect
    @property
    def on_disconnect(self) -> OptionalControlEventCallable:
        return self.__on_disconnect.handler

    @on_disconnect.setter
    def on_disconnect(self, handler: OptionalControlEventCallable):
        self.__on_disconnect.handler = handler

    # on_login
    @property
    def on_login(self) -> OptionalEventCallable["LoginEvent"]:
        return self.__on_login.handler

    @on_login.setter
    def on_login(self, handler: OptionalEventCallable["LoginEvent"]):
        self.__on_login.handler = handler

    # on_logout
    @property
    def on_logout(self) -> OptionalControlEventCallable:
        return self.__on_logout.handler

    @on_logout.setter
    def on_logout(self, handler: OptionalControlEventCallable):
        self.__on_logout.handler = handler

    # on_error
    @property
    def on_error(self) -> OptionalControlEventCallable:
        return self.__on_error.handler

    @on_error.setter
    def on_error(self, handler: OptionalControlEventCallable):
        self.__on_error.handler = handler

    # on_scroll
    @property
    def on_scroll(self):
        return self.__default_view.on_scroll

    @on_scroll.setter
    def on_scroll(self, handler: Optional[Callable[[OnScrollEvent], None]]):
        self.__default_view.on_scroll = handler

    # Magic methods
    def __contains__(self, item: Control) -> bool:
        return item in self._controls


class Offstage(Control):
    def __init__(
        self,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
    ) -> None:
        Control.__init__(
            self,
            visible=visible,
            disabled=disabled,
            data=data,
        )

        self.__controls: List[Control] = []
        self.__banner = None
        self.__snack_bar = None
        self.__dialog = None
        self.__bottom_sheet = None
        self.__splash = None

    def _get_control_name(self):
        return "offstage"

    def _get_children(self):
        return self.__controls

    # controls
    @property
    def controls(self) -> List[Control]:
        return self.__controls

    # splash
    @property
    def splash(self) -> Optional[Control]:
        return self.__splash

    @splash.setter
    def splash(self, value: Optional[Control]):
        self.__splash = value
        if value is not None:
            self.__controls.append(value)

    # banner
    @property
    def banner(self) -> Optional[Banner]:
        return self.__banner

    @banner.setter
    def banner(self, value: Optional[Banner]):
        self.__banner = value
        if value is not None:
            self.__controls.append(value)

    # snack_bar
    @property
    def snack_bar(self) -> Optional[SnackBar]:
        return self.__snack_bar

    @snack_bar.setter
    def snack_bar(self, value: Optional[SnackBar]):
        self.__snack_bar = value
        if value is not None:
            self.__controls.append(value)

    # dialog
    @property
    def dialog(self) -> Optional[Union[AlertDialog, CupertinoAlertDialog]]:
        return self.__dialog

    @dialog.setter
    def dialog(self, value: Optional[Union[AlertDialog, CupertinoAlertDialog]]):
        self.__dialog = value
        if value is not None:
            self.__controls.append(value)

    # bottom_sheet
    @property
    def bottom_sheet(self) -> Optional[Union[BottomSheet, CupertinoBottomSheet]]:
        return self.__bottom_sheet

    @bottom_sheet.setter
    def bottom_sheet(self, value: Optional[Union[BottomSheet, CupertinoBottomSheet]]):
        self.__bottom_sheet = value
        if value is not None:
            self.__controls.append(value)


@dataclass
class RouteChangeEvent(ControlEvent):
    route: str


@dataclass
class ViewPopEvent(ControlEvent):
    view: View


@dataclass
class KeyboardEvent(ControlEvent):
    key: str
    shift: bool
    ctrl: bool
    alt: bool
    meta: bool


class LoginEvent(ControlEvent):
    def __init__(
        self,
        error: str,
        error_description: str,
        target: str,
        name: str,
        data: str,
        control,
        page,
    ) -> None:
        super().__init__(target, name, data, control, page)

        self.error = error
        self.error_description = error_description


@dataclass
class InvokeMethodResults:
    method_id: str
    result: Optional[str]
    error: Optional[str]


class PageMediaData(ControlEvent):
    def __init__(self, padding, view_padding, view_insets) -> None:
        self.padding = Padding(
            left=padding["left"],
            top=padding["top"],
            right=padding["right"],
            bottom=padding["bottom"],
        )
        self.view_padding = Padding(
            left=view_padding["left"],
            top=view_padding["top"],
            right=view_padding["right"],
            bottom=view_padding["bottom"],
        )
        self.view_insets = Padding(
            left=view_insets["left"],
            top=view_insets["top"],
            right=view_insets["right"],
            bottom=view_insets["bottom"],
        )

    def __str__(self) -> str:
        return f"PageMediaData(padding={self.padding}, view_padding={self.view_padding}, view_insets={self.view_insets})"


class AppLifecycleStateChangeEvent(ControlEvent):
    def __init__(self, e: ControlEvent) -> None:
        super().__init__(e.target, e.name, e.data, e.control, e.page)
        self.state = AppLifecycleState(e.data)


class WindowEvent(ControlEvent):
    def __init__(self, e: ControlEvent):
        super().__init__(e.target, e.name, e.data, e.control, e.page)
        self.type = WindowEventType(e.data)


class WindowResizeEvent(ControlEvent):
    def __init__(self, e: ControlEvent):
        super().__init__(e.target, e.name, e.data, e.control, e.page)
        size = json.loads(e.data)
        self.width: float = size["width"]
        self.height: float = size["height"]



from enum import Enum
from typing import Any, Optional, Union

from flet.core.adaptive_control import AdaptiveControl
from flet.core.animation import AnimationValue
from flet.core.badge import BadgeValue
from flet.core.buttons import OutlinedBorder
from flet.core.constrained_control import ConstrainedControl
from flet.core.control import Control, OptionalNumber
from flet.core.ref import Ref
from flet.core.tooltip import TooltipValue
from flet.core.types import (
    ClipBehavior,
    ColorEnums,
    ColorValue,
    MarginValue,
    OffsetValue,
    OptionalControlEventCallable,
    ResponsiveNumber,
    RotateValue,
    ScaleValue,
)


class CardVariant(Enum):
    ELEVATED = "elevated"
    FILLED = "filled"
    OUTLINED = "outlined"


class Card(ConstrainedControl, AdaptiveControl):
    """
    A material design card: a panel with slightly rounded corners and an elevation shadow.

    Example:
    ```
    import flet as ft

    def main(page):
        page.title = "Card Example"
        page.add(
            ft.Card(
                content=ft.Container(
                    content=ft.Column(
                        [
                            ft.ListTile(
                                leading=ft.Icon(ft.icons.ALBUM),
                                title=ft.Text("The Enchanted Nightingale"),
                                subtitle=ft.Text(
                                    "Music by Julie Gable. Lyrics by Sidney Stein."
                                ),
                            ),
                            ft.Row(
                                [ft.TextButton("Buy tickets"), ft.TextButton("Listen")],
                                alignment=ft.MainAxisAlignment.END,
                            ),
                        ]
                    ),
                    width=400,
                    padding=10,
                )
            )
        )

    ft.app(target=main)

    ```

    -----

    Online docs: https://flet.dev/docs/controls/card
    """

    def __init__(
        self,
        content: Optional[Control] = None,
        margin: Optional[MarginValue] = None,
        elevation: OptionalNumber = None,
        color: Optional[ColorValue] = None,
        shadow_color: Optional[ColorValue] = None,
        surface_tint_color: Optional[ColorValue] = None,
        shape: Optional[OutlinedBorder] = None,
        clip_behavior: Optional[ClipBehavior] = None,
        is_semantic_container: Optional[bool] = None,
        show_border_on_foreground: Optional[bool] = None,
        variant: Optional[CardVariant] = None,
        #
        # ConstrainedControl and AdaptiveControl
        #
        ref: Optional[Ref] = None,
        width: OptionalNumber = None,
        height: OptionalNumber = None,
        left: OptionalNumber = None,
        top: OptionalNumber = None,
        right: OptionalNumber = None,
        bottom: OptionalNumber = None,
        expand: Union[None, bool, int] = None,
        expand_loose: Optional[bool] = None,
        col: Optional[ResponsiveNumber] = None,
        opacity: OptionalNumber = None,
        rotate: Optional[RotateValue] = None,
        scale: Optional[ScaleValue] = None,
        offset: Optional[OffsetValue] = None,
        aspect_ratio: OptionalNumber = None,
        animate_opacity: Optional[AnimationValue] = None,
        animate_size: Optional[AnimationValue] = None,
        animate_position: Optional[AnimationValue] = None,
        animate_rotation: Optional[AnimationValue] = None,
        animate_scale: Optional[AnimationValue] = None,
        animate_offset: Optional[AnimationValue] = None,
        on_animation_end: OptionalControlEventCallable = None,
        tooltip: Optional[TooltipValue] = None,
        badge: Optional[BadgeValue] = None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
        key: Optional[str] = None,
        adaptive: Optional[bool] = None,
    ):
        ConstrainedControl.__init__(
            self,
            ref=ref,
            key=key,
            width=width,
            height=height,
            left=left,
            top=top,
            right=right,
            bottom=bottom,
            expand=expand,
            expand_loose=expand_loose,
            col=col,
            opacity=opacity,
            rotate=rotate,
            scale=scale,
            offset=offset,
            aspect_ratio=aspect_ratio,
            animate_opacity=animate_opacity,
            animate_size=animate_size,
            animate_position=animate_position,
            animate_rotation=animate_rotation,
            animate_scale=animate_scale,
            animate_offset=animate_offset,
            on_animation_end=on_animation_end,
            tooltip=tooltip,
            badge=badge,
            visible=visible,
            disabled=disabled,
            data=data,
        )

        AdaptiveControl.__init__(self, adaptive=adaptive)

        self.content = content
        self.margin = margin
        self.elevation = elevation
        self.color = color
        self.shadow_color = shadow_color
        self.surface_tint_color = surface_tint_color
        self.shape = shape
        self.clip_behavior = clip_behavior
        self.is_semantic_container = is_semantic_container
        self.show_border_on_foreground = show_border_on_foreground
        self.variant = variant

    def _get_control_name(self):
        return "card"

    def before_update(self):
        super().before_update()
        self._set_attr_json("margin", self.__margin)
        self._set_attr_json("shape", self.__shape)

    def _get_children(self):
        children = []
        if self.__content is not None:
            self.__content._set_attr_internal("n", "content")
            children.append(self.__content)
        return children

    # margin
    @property
    def margin(self) -> Optional[MarginValue]:
        return self.__margin

    @margin.setter
    def margin(self, value: Optional[MarginValue]):
        self.__margin = value

    # elevation
    @property
    def elevation(self) -> OptionalNumber:
        return self._get_attr("elevation")

    @elevation.setter
    def elevation(self, value: OptionalNumber):
        self._set_attr("elevation", value)

    # color
    @property
    def color(self):
        return self.__color

    @color.setter
    def color(self, value):
        self.__color = value
        self._set_enum_attr("color", value, ColorEnums)

    # shadow_color
    @property
    def shadow_color(self):
        return self.__shadow_color

    @shadow_color.setter
    def shadow_color(self, value):
        self.__shadow_color = value
        self._set_enum_attr("shadowColor", value, ColorEnums)

    # surface_tint_color
    @property
    def surface_tint_color(self):
        return self.__surface_tint_color

    @surface_tint_color.setter
    def surface_tint_color(self, value):
        self.__surface_tint_color = value
        self._set_enum_attr("surfaceTintColor", value, ColorEnums)

    # shape
    @property
    def shape(self) -> Optional[OutlinedBorder]:
        return self.__shape

    @shape.setter
    def shape(self, value: Optional[OutlinedBorder]):
        self.__shape = value

    # content
    @property
    def content(self) -> Optional[Control]:
        return self.__content

    @content.setter
    def content(self, value: Optional[Control]):
        self.__content = value

    # clip_behavior
    @property
    def clip_behavior(self) -> Optional[ClipBehavior]:
        return self.__clip_behavior

    @clip_behavior.setter
    def clip_behavior(self, value: Optional[ClipBehavior]):
        self.__clip_behavior = value
        self._set_enum_attr("clipBehavior", value, ClipBehavior)

    # is_semantic_container
    @property
    def is_semantic_container(self) -> bool:
        return self._get_attr("isSemanticContainer", data_type="bool", def_value=True)

    @is_semantic_container.setter
    def is_semantic_container(self, value):
        self._set_attr("isSemanticContainer", value)

    # show_border_on_foreground
    @property
    def show_border_on_foreground(self) -> bool:
        return self._get_attr(
            "showBorderOnForeground", data_type="bool", def_value=True
        )

    @show_border_on_foreground.setter
    def show_border_on_foreground(self, value):
        self._set_attr("showBorderOnForeground", value)

    # variant
    @property
    def variant(self) -> Optional[CardVariant]:
        return self.__variant

    @variant.setter
    def variant(self, value: Optional[CardVariant]):
        self.__variant = value
        self._set_enum_attr("variant", value, CardVariant)



from dataclasses import dataclass
from typing import Optional, Union

from flet.core.alignment import Alignment
from flet.core.text_style import TextStyle
from flet.core.types import ColorValue, OffsetValue, OptionalNumber, PaddingValue


@dataclass
class Badge:
    """Badges are used to show notifications, counts, or status information on navigation items such as NavigationBar or NavigationRail destinations
    or a button's icon."""

    text: Optional[str] = None
    offset: Optional[OffsetValue] = None
    alignment: Optional[Alignment] = None
    bgcolor: Optional[ColorValue] = None
    label_visible: Optional[bool] = None
    large_size: OptionalNumber = None
    padding: Optional[PaddingValue] = None
    small_size: OptionalNumber = None
    text_color: Optional[ColorValue] = None
    text_style: Optional[TextStyle] = None


BadgeValue = Union[str, "Badge"]



from typing import Any, Optional

from flet.core.border import BorderSide
from flet.core.control import Control, OptionalNumber
from flet.core.ref import Ref
from flet.core.text_style import TextStyle
from flet.core.types import ColorEnums, ColorValue


class PieChartSection(Control):
    def __init__(
        self,
        value: OptionalNumber = None,
        radius: OptionalNumber = None,
        color: Optional[ColorValue] = None,
        border_side: Optional[BorderSide] = None,
        title: Optional[str] = None,
        title_style: Optional[TextStyle] = None,
        title_position: OptionalNumber = None,
        badge: Optional[Control] = None,
        badge_position: OptionalNumber = None,
        #
        # Control
        #
        ref: Optional[Ref] = None,
        disabled: Optional[bool] = None,
        visible: Optional[bool] = None,
        data: Any = None,
    ):

        Control.__init__(
            self,
            ref=ref,
            disabled=disabled,
            visible=visible,
            data=data,
        )

        self.value = value
        self.radius = radius
        self.color = color
        self.border_side = border_side
        self.title = title
        self.title_style = title_style
        self.title_position = title_position
        self.badge = badge
        self.badge_position = badge_position

    def _get_control_name(self):
        return "section"

    def before_update(self):
        super().before_update()
        self._set_attr_json("borderSide", self.__border_side)
        self._set_attr_json("titleStyle", self.__title_style)

    def _get_children(self):
        children = []
        if self.__badge:
            self.__badge._set_attr_internal("n", "badge")
            children.append(self.__badge)
        return children

    # value
    @property
    def value(self) -> OptionalNumber:
        return self._get_attr("value", data_type="float")

    @value.setter
    def value(self, value: OptionalNumber):
        self._set_attr("value", value)

    # radius
    @property
    def radius(self) -> OptionalNumber:
        return self._get_attr("radius", data_type="float")

    @radius.setter
    def radius(self, value: OptionalNumber):
        self._set_attr("radius", value)

    # border_side
    @property
    def border_side(self) -> Optional[BorderSide]:
        return self.__border_side

    @border_side.setter
    def border_side(self, value: Optional[BorderSide]):
        self.__border_side = value

    # color
    @property
    def color(self) -> Optional[ColorValue]:
        return self.__color

    @color.setter
    def color(self, value: Optional[ColorValue]):
        self.__color = value
        self._set_enum_attr("color", value, ColorEnums)

    # badge
    @property
    def badge(self) -> Optional[Control]:
        return self.__badge

    @badge.setter
    def badge(self, value: Optional[Control]):
        self.__badge = value

    # badge_position
    @property
    def badge_position(self) -> OptionalNumber:
        return self._get_attr("badgePosition", data_type="float")

    @badge_position.setter
    def badge_position(self, value: OptionalNumber):
        self._set_attr("badgePosition", value)

    # title
    @property
    def title(self):
        return self._get_attr("title")

    @title.setter
    def title(self, value: Optional[str]):
        self._set_attr("title", value)

    # title_style
    @property
    def title_style(self):
        return self.__title_style

    @title_style.setter
    def title_style(self, value: Optional[TextStyle]):
        self.__title_style = value

    # title_position
    @property
    def title_position(self) -> float:
        return self._get_attr("titlePosition", data_type="float", def_value=1.0)

    @title_position.setter
    def title_position(self, value: OptionalNumber):
        self._set_attr("titlePosition", value)



from typing import Any, List, Optional

from flet.core.canvas.path import Path
from flet.core.canvas.shape import Shape
from flet.core.control import OptionalNumber
from flet.core.types import ColorEnums, ColorValue


class Shadow(Shape):
    def __init__(
        self,
        path: Optional[List[Path.PathElement]] = None,
        color: Optional[ColorValue] = None,
        elevation: OptionalNumber = None,
        transparent_occluder: Optional[bool] = None,
        #
        # Control
        #
        ref=None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
    ):
        Shape.__init__(self, ref=ref, visible=visible, disabled=disabled, data=data)

        self.path = path
        self.color = color
        self.elevation = elevation
        self.transparent_occluder = transparent_occluder

    def _get_control_name(self):
        return "shadow"

    def before_update(self):
        super().before_update()
        self._set_attr_json("path", self.__path)

    # path
    @property
    def path(self):
        return self.__path

    @path.setter
    def path(self, value: Optional[List[Path.PathElement]]):
        self.__path = value if value is not None else []

    # color
    @property
    def color(self) -> Optional[ColorValue]:
        return self.__color

    @color.setter
    def color(self, value: Optional[ColorValue]):
        self.__color = value
        self._set_enum_attr("color", value, ColorEnums)

    # elevation
    @property
    def elevation(self) -> OptionalNumber:
        return self._get_attr("elevation")

    @elevation.setter
    def elevation(self, value: OptionalNumber):
        self._set_attr("elevation", value)

    # transparent_occluder
    @property
    def transparent_occluder(self) -> bool:
        return self._get_attr("transparentOccluder", data_type="bool", def_value=False)

    @transparent_occluder.setter
    def transparent_occluder(self, value: Optional[bool]):
        self._set_attr("transparentOccluder", value)
